"""Kat parser to convert tokens to productions.

This is a recursive descent parser, providing unlimited lookahead capabilities to allow
arbitrary context-free grammars to be parsed. The implementation of memoization makes it
a *packrat* parser which runs in linear time at the expense of potentially unlimited
memory use. In practice memory use is limited by the simplicity of typical kat scripts
(e.g. expressions don't tend to have many subexpressions).

This is inspired by Python's PEG parser, used as of 3.9. Some introductory information
can be found in `this series of blog posts
<https://medium.com/@gvanrossum_83706/peg-parsing-series-de5d41b2ed60>`__.

Sean Leavey <sean.leavey@ligo.org>
"""

import logging
from io import StringIO
from .containers import (
    KatScript,
    KatElement,
    KatFunction,
    KatKwarg,
    KatExpression,
    KatGroupedExpression,
    KatArray,
    KatNumericalArray,
)
from .tokenizer import KatTokenizer
from .memoize import memoize, memoize_left_rec
from .exceptions import KatSyntaxError, KatMissingAfterDirective

LOGGER = logging.getLogger(__name__)


class KatParser:
    """Kat script parser.

    This uses so-called *packrat* parsing to reduce, via productions, tokens yielded
    from a token stream generated from an input file or string to
    :class:`.KatScriptItem` objects containing the associated :class:`tokens
    <.KatToken>`.
    """

    def __init__(self):
        self.tokens = None
        self.pos = None
        self.memos = None
        self._tokenizer = None
        self._token_stream = None

    @property
    def script(self):
        return self._tokenizer.script

    def parse(self, string):
        """Parse the contents of `string`.

        Parameters
        ----------
        string : :class:`str`
            The string to parse kat script from.

        Returns
        -------
        :class:`.KatScript`
            The parsed kat script.
        """
        return self.parse_file(StringIO(string))

    def parse_file(self, fobj):
        """Parse the contents of the specified file.

        Parameters
        ----------
        fobj : :class:`io.FileIO`
            The file object to parse kat script from. This should be opened in text
            mode.

        Returns
        -------
        :class:`.KatScript`
            The parsed kat script.
        """
        # Reset parser state.
        self.tokens = []
        self.pos = 0
        self.memos = {}
        self._log_stack = []

        # Perform parse.
        self._tokenizer = KatTokenizer()
        self._token_stream = self._tokenizer.tokenize_file(fobj)

        if (script := self.expect_production("start")) is not None:
            return script

        # There was an error.
        self._diagnose_error()

    def _diagnose_error(self):
        if not self.tokens:
            error_token = self.peek_token()
        else:
            error_token = self.tokens[-1]

        raise KatSyntaxError("syntax error", self.script, error_token)

    def mark(self):
        return self.pos

    def reset(self, pos):
        if pos == self.pos:
            return
        self.pos = pos

    def get_token(self):
        token = self.peek_token()
        self.pos += 1
        return token

    def peek_token(self):
        if self.pos == len(self.tokens):
            self.tokens.append(next(self._token_stream))
        return self.tokens[self.pos]

    def positive_lookahead(self, token_type):
        pos = self.mark()
        token = self.peek_token()
        self.reset(pos)
        return token.type == token_type

    def negative_lookahead(self, token_type):
        return not self.positive_lookahead(token_type)

    def maybe_whitespace(self, multiline, whitespace):
        """Expect zero or more whitespace if enabled, else return []."""
        if not whitespace:
            return []
        return self.loop("empty", False, multiline)

    def maybe_trailing_comma(self, multiline):
        pos = self.mark()
        if (
            True
            and (empty := self.loop("empty", False, multiline)) is not None
            and (COMMA := self.expect_token("COMMA")) is not None
        ):
            return [*empty, COMMA]
        self.reset(pos)
        return []

    @memoize
    def expect_token(self, arg):
        self._log_stack.append(arg)
        path = "->".join(self._log_stack)
        # path = arg
        LOGGER.debug(f"{path}?")
        token = self.peek_token()

        if token.type == arg:
            LOGGER.debug(f"{path} = {token!r}!")
            result = self.get_token()
            self._log_stack.pop()
            return result

        LOGGER.debug(f"{arg} not found")
        self._log_stack.pop()

    def expect_production(self, production, *args, **kwargs):
        self._log_stack.append(production)
        path = "->".join(self._log_stack)
        # path = production
        LOGGER.debug(f"{path}?")

        result = getattr(self, production)(*args, **kwargs)

        if result is not None:
            LOGGER.debug(f"{path} = {result!r}!")
            self._log_stack.pop()
            return result

        LOGGER.debug(f"{production} not found")

        self._log_stack.pop()

    def loop(self, production, nonempty, *args, **kwargs):
        mark = self.mark()
        nodes = []
        while (node := self.expect_production(production, *args, **kwargs)) is not None:
            nodes.append(node)
        if len(nodes) >= nonempty:
            return nodes
        self.reset(mark)

    def start(self):
        pos = self.mark()

        if (
            True
            and (script_lines := self.loop("script_line", True)) is not None
            and self.expect_token("ENDMARKER") is not None
        ):
            statements = []
            extra = []
            for statement, line_extra in script_lines:
                if statement:
                    statements.append(statement)
                extra.extend(line_extra)

            return KatScript(arguments=statements, extra=extra)

        self.reset(pos)

        # Empty file.
        if self.expect_token("ENDMARKER") is not None:
            return KatScript(arguments=[], extra=[])

        self.reset(pos)

    @memoize
    def script_line(self):
        pos = self.mark()

        # script_line -> script_line_empty* statement script_line_empty* NEWLINE
        if (
            True
            and (script_line_empty1 := self.loop("script_line_empty", False))
            is not None
            and (statement := self.expect_production("statement")) is not None
            and (script_line_empty2 := self.loop("script_line_empty", False))
            is not None
            and (NEWLINE := self.expect_token("NEWLINE")) is not None
        ):
            extra = [*script_line_empty1, *script_line_empty2]
            # Ignore implicit newlines, which are metatokens.
            if hasattr(NEWLINE, "value"):
                extra.append(NEWLINE)
            return statement, extra

        self.reset(pos)

        # script_line -> script_line_empty* NEWLINE
        if (
            True
            and (script_line_empty := self.loop("script_line_empty", False)) is not None
            and (NEWLINE := self.expect_token("NEWLINE")) is not None
        ):
            extra = [*script_line_empty]
            if hasattr(NEWLINE, "value"):
                # Newline is a real token.
                extra.append(NEWLINE)
            return None, [*script_line_empty, NEWLINE]

        self.reset(pos)

        if (error := self.expect_production("invalid_script_line")) is not None:
            raise error

    @memoize
    def script_line_empty(self):
        pos = self.mark()

        # script_line_empty -> WHITESPACE
        if (WHITESPACE := self.expect_token("WHITESPACE")) is not None:
            return WHITESPACE

        self.reset(pos)

        # script_line_empty -> COMMENT
        if (COMMENT := self.expect_token("COMMENT")) is not None:
            return COMMENT

        self.reset(pos)

    @memoize
    def empty(self, multiline):
        pos = self.mark()

        # empty -> WHITESPACE
        if (WHITESPACE := self.expect_token("WHITESPACE")) is not None:
            return WHITESPACE

        self.reset(pos)

        if multiline:
            # empty -> COMMENT
            if (COMMENT := self.expect_token("COMMENT")) is not None:
                return COMMENT

            self.reset(pos)

            # empty -> NEWLINE
            if (NEWLINE := self.expect_token("NEWLINE")) is not None:
                return NEWLINE

            self.reset(pos)

    @memoize
    def statement(self):
        pos = self.mark()

        # statement -> element
        if (element := self.expect_production("element")) is not None:
            return element

        self.reset(pos)

        # statement -> function
        if (function := self.expect_production("function", True)) is not None:
            return function

        self.reset(pos)

    @memoize
    def element(self):
        pos = self.mark()

        # element -> NAME WHITESPACE NAME WHITESPACE element_params
        if (
            True
            and (DIRECTIVE := self.expect_token("NAME")) is not None
            and (WHITESPACE1 := self.expect_token("WHITESPACE")) is not None
            and (NAME := self.expect_token("NAME")) is not None
            and (WHITESPACE2 := self.expect_token("WHITESPACE")) is not None
            and (element_params := self.expect_production("element_params")) is not None
        ):
            arguments, params_extra = element_params
            return KatElement(
                directive=DIRECTIVE,
                name=NAME,
                arguments=arguments,
                extra=[WHITESPACE1, WHITESPACE2, *params_extra],
            )

        self.reset(pos)

        # element -> NAME WHITESPACE NAME
        if (
            True
            and (DIRECTIVE := self.expect_token("NAME")) is not None
            and (WHITESPACE := self.expect_token("WHITESPACE")) is not None
            and (NAME := self.expect_token("NAME")) is not None
        ):
            return KatElement(
                directive=DIRECTIVE, name=NAME, arguments=[], extra=[WHITESPACE],
            )

        self.reset(pos)

    @memoize
    def element_params(self):
        pos = self.mark()

        # element_params -> element_value_list empty+ element_key_value_list
        if (
            True
            and (element_value_list := self.expect_production("element_value_list"))
            is not None
            and (empty := self.loop("empty", True, False)) is not None
            and (
                element_key_value_list := self.expect_production(
                    "element_key_value_list"
                )
            )
            is not None
        ):
            values, value_extra = element_value_list
            key_values, key_value_extra = element_key_value_list
            return ([*values, *key_values], [*value_extra, *empty, *key_value_extra])

        self.reset(pos)

        # element_params -> element_value_list
        if (
            element_value_list := self.expect_production("element_value_list")
        ) is not None:
            return element_value_list

        self.reset(pos)

        # element_params -> element_key_value_list
        if (
            element_key_value_list := self.expect_production("element_key_value_list")
        ) is not None:
            return element_key_value_list

        self.reset(pos)

    @memoize
    def element_value_list(self):
        pos = self.mark()

        # element_value_list -> positional_value empty+ element_value_list
        if (
            True
            and (
                positional_value := self.expect_production(
                    "positional_value", False, False
                )
            )
            is not None
            and (empty := self.loop("empty", True, False)) is not None
            and (element_value_list := self.expect_production("element_value_list"))
            is not None
        ):
            values, value_extra = element_value_list
            return ([positional_value, *values], [*empty, *value_extra])

        self.reset(pos)

        # element_value_list -> positional_value
        if (
            positional_value := self.expect_production("positional_value", False, False)
        ) is not None:
            return [positional_value], []

        self.reset(pos)

    @memoize_left_rec
    def element_key_value_list(self):
        pos = self.mark()

        # element_key_value_list -> element_key_value_list empty+ element_key_value_list
        if (
            True
            and (
                element_key_value_list1 := self.expect_production(
                    "element_key_value_list"
                )
            )
            is not None
            and (empty := self.loop("empty", True, False)) is not None
            and (
                element_key_value_list2 := self.expect_production(
                    "element_key_value_list"
                )
            )
            is not None
        ):
            key_values1, key_value_extra1 = element_key_value_list1
            key_values2, key_value_extra2 = element_key_value_list2
            return (
                [*key_values1, *key_values2],
                [*key_value_extra1, *empty, *key_value_extra2],
            )

        self.reset(pos)

        # element_key_value_list -> key_value
        if (key_value := self.expect_production("key_value", False, False)) is not None:
            return [key_value], []

        self.reset(pos)

    @memoize
    def function(self, multiline):
        pos = self.mark()

        # function -> NAME '(' empty* function_params empty* ')'
        if (
            True
            and (FUNCTION := self.expect_token("NAME")) is not None
            and (LPAREN := self.expect_token("LPAREN")) is not None
            and (empty1 := self.loop("empty", False, multiline)) is not None
            and (
                function_params := self.expect_production("function_params", multiline)
            )
            is not None
            and (empty2 := self.loop("empty", False, multiline)) is not None
            and (RPAREN := self.expect_token("RPAREN")) is not None
        ):
            args, extra = function_params
            return KatFunction(
                directive=FUNCTION,
                arguments=args,
                extra=[LPAREN, *empty1, *extra, *empty2, RPAREN],
            )

        self.reset(pos)

        # function -> NAME '(' empty* ')'
        if (
            True
            and (FUNCTION := self.expect_token("NAME")) is not None
            and (LPAREN := self.expect_token("LPAREN")) is not None
            and (empty := self.loop("empty", False, multiline)) is not None
            and (RPAREN := self.expect_token("RPAREN")) is not None
        ):
            return KatFunction(
                directive=FUNCTION, arguments=[], extra=[LPAREN, *empty, RPAREN]
            )

        self.reset(pos)

    @memoize
    def function_params(self, multiline):
        pos = self.mark()

        # function_params -> function_value_list empty* ',' empty* function_key_value_list (empty* ',')?
        if (
            True
            and (
                function_value_list := self.expect_production(
                    "function_value_list", multiline
                )
            )
            is not None
            and (empty1 := self.loop("empty", False, multiline)) is not None
            and (COMMA := self.expect_token("COMMA")) is not None
            and (empty2 := self.loop("empty", False, multiline)) is not None
            and (
                function_key_value_list := self.expect_production(
                    "function_key_value_list", multiline
                )
            )
            is not None
            and (trailing_comma := self.maybe_trailing_comma(multiline)) is not None
        ):
            values, value_extra = function_value_list
            key_values, key_value_extra = function_key_value_list
            return (
                [*values, *key_values],
                [
                    *value_extra,
                    *empty1,
                    COMMA,
                    *empty2,
                    *key_value_extra,
                    *trailing_comma,
                ],
            )

        self.reset(pos)

        # function_params -> function_value_list (empty* ',')?
        if (
            True
            and (
                function_value_list := self.expect_production(
                    "function_value_list", multiline
                )
            )
            is not None
            and (trailing_comma := self.maybe_trailing_comma(multiline)) is not None
        ):
            values, value_extra = function_value_list
            return values, [*value_extra, *trailing_comma]

        self.reset(pos)

        # function_params -> function_key_value_list (empty* ',')?
        if (
            True
            and (
                function_key_value_list := self.expect_production(
                    "function_key_value_list", multiline
                )
            )
            is not None
            and (trailing_comma := self.maybe_trailing_comma(multiline)) is not None
        ):
            key_values, key_value_extra = function_key_value_list
            return key_values, [*key_value_extra, *trailing_comma]

        self.reset(pos)

    @memoize
    def function_value_list(self, multiline):
        pos = self.mark()

        # function_value_list -> positional_value empty* ',' empty* function_value_list
        if (
            True
            and (
                positional_value := self.expect_production(
                    "positional_value", multiline, True
                )
            )
            is not None
            and (empty1 := self.loop("empty", False, multiline)) is not None
            and (COMMA1 := self.expect_token("COMMA")) is not None
            and (empty2 := self.loop("empty", False, multiline)) is not None
            and (
                function_value_list := self.expect_production(
                    "function_value_list", multiline
                )
            )
            is not None
        ):
            values, value_extra = function_value_list
            return (
                [positional_value, *values],
                [*empty1, COMMA1, *empty2, *value_extra],
            )

        self.reset(pos)

        # function_value_list -> positional_value
        if (
            positional_value := self.expect_production(
                "positional_value", multiline, True
            )
        ) is not None:
            return [positional_value], []

        self.reset(pos)

    @memoize_left_rec
    def function_key_value_list(self, multiline):
        pos = self.mark()

        # function_key_value_list -> function_key_value_list empty* ',' empty* function_key_value_list
        if (
            True
            and (
                function_key_value_list1 := self.expect_production(
                    "function_key_value_list", multiline
                )
            )
            is not None
            and (empty1 := self.loop("empty", False, multiline)) is not None
            and (COMMA := self.expect_token("COMMA")) is not None
            and (empty2 := self.loop("empty", False, multiline)) is not None
            and (
                function_key_value_list2 := self.expect_production(
                    "function_key_value_list", multiline
                )
            )
            is not None
        ):
            key_values1, key_value_extra1 = function_key_value_list1
            key_values2, key_value_extra2 = function_key_value_list2
            return (
                [*key_values1, *key_values2],
                [*key_value_extra1, *empty1, COMMA, *empty2, *key_value_extra2],
            )

        self.reset(pos)

        # function_key_value_list -> key_value
        if (
            key_value := self.expect_production("key_value", multiline, True)
        ) is not None:
            return [key_value], []

        self.reset(pos)

    @memoize
    def positional_value(self, multiline, whitespace):
        pos = self.mark()

        # Don't match values followed by '=', which are kwarg keys.
        if (
            True
            and (value := self.expect_production("value", multiline, whitespace))
            is not None
            and self.negative_lookahead("EQUALS")
        ):
            return value

        self.reset(pos)

    @memoize
    def value(self, multiline, whitespace):
        pos = self.mark()

        # value -> array
        # Should be above `expr` so arrays that aren't part of expressions stay as lists
        # instead of get converted into ndarrays.
        if (array := self.expect_production("array", multiline)) is not None:
            return array

        self.reset(pos)

        # value -> expr
        if (expr := self.expect_production("expr", multiline, whitespace)) is not None:
            return expr

        self.reset(pos)

        # value -> function
        if (function := self.expect_production("function", multiline)) is not None:
            return function

        self.reset(pos)

        # value -> NONE / BOOLEAN / STRING
        for token_type in ("NONE", "BOOLEAN", "STRING"):
            if (TOKEN := self.expect_token(token_type)) is not None:
                return TOKEN

            self.reset(pos)

    @memoize
    def key_value(self, multiline, whitespace):
        pos = self.mark()

        # key_value -> NAME '=' value
        if (
            True
            and (NAME := self.expect_token("NAME")) is not None
            and (EQUALS := self.expect_token("EQUALS")) is not None
            and (value := self.expect_production("value", multiline, whitespace))
            is not None
        ):
            return KatKwarg(key=NAME, value=value, extra=[EQUALS])

        self.reset(pos)

        if (
            True
            and (NAME := self.expect_token("NAME")) is not None
            and (EQUALS := self.expect_token("EQUALS")) is not None
            and self.expect_production("value", multiline, whitespace) is None  # !
        ):
            raise KatSyntaxError("missing value", self.script, EQUALS)

        self.reset(pos)

    @memoize_left_rec
    def expr(self, multiline, whitespace):
        pos = self.mark()

        # NOTE: operator precedence is set by defining productions within productions.
        # expr -> expr ( '+' / '-' ) expr1
        for operator in ("PLUS", "MINUS"):
            if (
                True
                and (lhs := self.expect_production("expr", multiline, whitespace))
                is not None
                and (whitespace1 := self.maybe_whitespace(multiline, whitespace))
                is not None
                and (operator := self.expect_token(operator)) is not None
                and (whitespace2 := self.maybe_whitespace(multiline, whitespace))
                is not None
                and (rhs := self.expect_production("expr1", multiline, whitespace))
                is not None
            ):
                return KatExpression(
                    operator=operator,
                    arguments=[lhs, rhs],
                    extra=[*whitespace1, *whitespace2],
                )

            self.reset(pos)

        # expr -> expr1
        if (
            expr1 := self.expect_production("expr1", multiline, whitespace)
        ) is not None:
            return expr1

        self.reset(pos)

    @memoize_left_rec
    def expr1(self, multiline, whitespace):
        """Times, divide and floordivide operators."""
        pos = self.mark()

        # expr1 -> expr1 ( '*' / '/' / '//' ) expr2
        for operator in ("TIMES", "DIVIDE", "FLOORDIVIDE"):
            if (
                True
                and (lhs := self.expect_production("expr1", multiline, whitespace))
                is not None
                and (whitespace1 := self.maybe_whitespace(multiline, whitespace))
                is not None
                and (operator := self.expect_token(operator)) is not None
                and (whitespace2 := self.maybe_whitespace(multiline, whitespace))
                is not None
                and (rhs := self.expect_production("expr2", multiline, whitespace))
                is not None
            ):
                return KatExpression(
                    operator=operator,
                    arguments=[lhs, rhs],
                    extra=[*whitespace1, *whitespace2],
                )

            self.reset(pos)

        # expr1 -> expr2
        if (
            expr2 := self.expect_production("expr2", multiline, whitespace)
        ) is not None:
            return expr2

        self.reset(pos)

    @memoize
    def expr2(self, multiline, whitespace):
        """Unary operators."""
        pos = self.mark()

        # expr2 -> ( '+' / '-' ) expr2
        for unary_operator in ("PLUS", "MINUS"):
            if (
                True
                and (unary_operator := self.expect_token(unary_operator)) is not None
                and (expr2 := self.expect_production("expr2", multiline, whitespace))
                is not None
            ):
                return KatFunction(
                    directive=unary_operator, arguments=[expr2], extra=[],
                )

            self.reset(pos)

        # expr2 -> expr3
        if (
            expr3 := self.expect_production("expr3", multiline, whitespace)
        ) is not None:
            return expr3

        self.reset(pos)

    @memoize
    def expr3(self, multiline, whitespace):
        """Power operator."""
        pos = self.mark()

        # expr3 -> expr4 '**' expr2
        if (
            True
            and (expr4 := self.expect_production("expr4", multiline, whitespace))
            is not None
            and (whitespace1 := self.maybe_whitespace(multiline, whitespace))
            is not None
            and (power := self.expect_token("POWER")) is not None
            and (whitespace2 := self.maybe_whitespace(multiline, whitespace))
            is not None
            and (expr2 := self.expect_production("expr2", multiline, whitespace))
            is not None
        ):
            return KatExpression(
                operator=power,
                arguments=[expr4, expr2],
                extra=[*whitespace1, *whitespace2],
            )

        self.reset(pos)

        # expr3 -> expr4
        if (
            expr4 := self.expect_production("expr4", multiline, whitespace)
        ) is not None:
            return expr4

        self.reset(pos)

    @memoize
    def expr4(self, multiline, whitespace):
        """Parentheses, functions, references, names and numbers.

        Names are allowed here to support keywords and copy-/read-by-value parameters
        like l1.P.
        """
        pos = self.mark()

        # expr4 -> '(' empty* expr empty* ')'
        if (
            True
            and (LPAREN := self.expect_token("LPAREN")) is not None
            and (whitespace1 := self.maybe_whitespace(multiline, True)) is not None
            and (expr := self.expect_production("expr", multiline, True)) is not None
            and (whitespace2 := self.maybe_whitespace(multiline, True)) is not None
            and (RPAREN := self.expect_token("RPAREN")) is not None
        ):
            return KatGroupedExpression(
                arguments=[expr], extra=[LPAREN, *whitespace1, *whitespace2, RPAREN]
            )

        self.reset(pos)

        # expr4 -> function
        if (function := self.expect_production("function", multiline)) is not None:
            return function

        self.reset(pos)

        # expr4 -> array
        if (array := self.expect_production("array", multiline)) is not None:
            LOGGER.debug("converting array expression operand to numerical array")
            return KatNumericalArray.from_array(array)

        self.reset(pos)

        # expr4 -> NUMBER / NAME / REFERENCE
        for token_type in ("NUMBER", "NAME", "REFERENCE"):
            if (TOKEN := self.expect_token(token_type)) is not None:
                return TOKEN

            self.reset(pos)

    @memoize
    def array(self, multiline):
        pos = self.mark()

        # array -> '[' empty* array_values empty* ']'
        if (
            True
            and (LBRACKET := self.expect_token("LBRACKET")) is not None
            and (empty1 := self.loop("empty", False, False)) is not None
            and (array_values := self.expect_production("array_values", multiline))
            is not None
            and (empty2 := self.loop("empty", False, False)) is not None
            and (RBRACKET := self.expect_token("RBRACKET")) is not None
        ):
            values, value_extra = array_values
            return KatArray(
                arguments=values,
                extra=[LBRACKET, *empty1, *value_extra, *empty1, *empty2, RBRACKET],
            )

        self.reset(pos)

        # array -> '[' empty* ']'
        if (
            True
            and (LBRACKET := self.expect_token("LBRACKET")) is not None
            and (empty := self.loop("empty", False, False)) is not None
            and (RBRACKET := self.expect_token("RBRACKET")) is not None
        ):
            return KatArray(arguments=[], extra=[LBRACKET, *empty, RBRACKET])

        self.reset(pos)

        if (error := self.expect_production("invalid_array", multiline)) is not None:
            raise error

        self.reset(pos)

    @memoize
    def array_values(self, multiline):
        pos = self.mark()

        # array_values -> array_value empty* ',' empty* array_values (empty* ',')?
        if (
            True
            and (array_value := self.expect_production("array_value", multiline))
            is not None
            and (empty1 := self.loop("empty", False, multiline)) is not None
            and (COMMA := self.expect_token("COMMA")) is not None
            and (empty2 := self.loop("empty", False, multiline)) is not None
            and (array_values := self.expect_production("array_values", multiline))
            is not None
            and (trailing_comma := self.maybe_trailing_comma(multiline)) is not None
        ):
            values, value_extra = array_values
            return (
                [array_value, *values],
                [*empty1, COMMA, *empty2, *value_extra, *trailing_comma],
            )

        self.reset(pos)

        # array_values -> array_value (empty* ',')?
        if (
            True
            and (array_value := self.expect_production("array_value", multiline))
            is not None
            and (trailing_comma := self.maybe_trailing_comma(multiline)) is not None
        ):
            return [array_value], trailing_comma

        self.reset(pos)

    @memoize
    def array_value(self, multiline):
        pos = self.mark()

        # array_value -> array
        # Should be above `expr` so arrays that aren't part of expressions stay as lists
        # instead of get converted into ndarrays.
        if (array := self.expect_production("array", multiline)) is not None:
            return array

        self.reset(pos)

        # array_value -> expr
        if (expr := self.expect_production("expr", multiline, True)) is not None:
            return expr

        self.reset(pos)

        # array_value -> NONE / NUMBER / NAME / BOOLEAN / REFERENCE / STRING
        for token_type in ("NONE", "NUMBER", "NAME", "BOOLEAN", "REFERENCE", "STRING"):
            if (TOKEN := self.expect_token(token_type)) is not None:
                return TOKEN

            self.reset(pos)

    @memoize
    def invalid_script_line(self):
        pos = self.mark()

        # Check for positional arguments specified after keyword arguments in inline
        # statements. Multiline statements are checked elsewhere.
        if (
            True
            and self.loop("script_line_empty", False) is not None
            and self.expect_production("statement") is not None
            and self.loop("script_line_empty", True) is not None
            and (element_value_list := self.expect_production("element_value_list"))
            is not None
        ):
            values, _ = element_value_list

            return KatSyntaxError(
                "positional argument follows keyword argument", self.script, values[0],
            )

        self.reset(pos)

        # Invalid parameter.
        # This has to be located here because we want the above rules to first check for
        # NEWLINE, to avoid matching valid elements.
        if (
            True
            and self.expect_token("NAME") is not None
            and self.expect_token("WHITESPACE") is not None
            and self.expect_token("NAME") is not None
            and self.expect_token("WHITESPACE") is not None
            and self.expect_production("element_params") is not None
            and (element_params := self.expect_production("element_params")) is not None
        ):
            arguments, _ = element_params
            return KatSyntaxError("syntax error", self.script, arguments[0])

        self.reset(pos)

        # Invalid element (no name) or function (no opening parenthesis).
        if (
            True
            and (DIRECTIVE := self.expect_token("NAME")) is not None
            and self.expect_token("WHITESPACE") is not None
            and self.expect_production("element_params") is not None
            and self.expect_token("NEWLINE") is not None
        ):
            return KatMissingAfterDirective(self.script, DIRECTIVE)

        self.reset(pos)

        # Invalid function: space before parentheses (non-empty parameters).
        if (
            True
            and self.expect_token("NAME") is not None
            and (WHITESPACE := self.expect_token("WHITESPACE")) is not None
            and self.expect_token("LPAREN") is not None
            and self.loop("empty", False, True) is not None
            and self.expect_production("function_params", True) is not None
            and self.loop("empty", False, True) is not None
            and self.expect_token("RPAREN") is not None
        ):
            return KatSyntaxError("space not allowed here", self.script, WHITESPACE)

        self.reset(pos)

        # Invalid function: space before parentheses (empty parameters).
        if (
            True
            and self.expect_token("NAME") is not None
            and (WHITESPACE := self.expect_token("WHITESPACE")) is not None
            and self.expect_token("LPAREN") is not None
            and self.loop("empty", False, True) is not None
            and self.expect_token("RPAREN") is not None
        ):
            return KatSyntaxError("space not allowed here", self.script, WHITESPACE)

        self.reset(pos)

    @memoize
    def invalid_array(self, multiline):
        pos = self.mark()

        # Invalid: no comma as a delimiter.
        if (
            True
            and self.expect_token("LBRACKET") is not None
            and self.expect_production("array_values", multiline) is not None
            and (empty := self.loop("empty", False, False)) is not None
            and self.expect_production("array_values", multiline) is not None
        ):
            return KatSyntaxError(
                "array values should be delimited by ','", self.script, empty[0]
            )

        self.reset(pos)



"""Kat file parser analysis (action tree) parsing fuzzing tests."""

from hypothesis import given, assume, settings
from finesse.script import parse
from ......util import kat_name
from ....util import dedent_multiline


@given(name=kat_name())
@settings(deadline=1000)  # See #292.
def test_component_name_fuzzing(name):
    parse(f"mirror {name}")


@given(name=kat_name())
@settings(deadline=1000)  # See #292.
def test_analysis_name_fuzzing(name):
    assume(name != "m1")
    parse(
        dedent_multiline(
            f"""
            mirror m1
            xaxis(m1.phi, lin, 0, 90, 100, name='{name}')
            """
        )
    )



"""Cavity unit tests."""

import pytest
from networkx.exception import NetworkXNoPath
from finesse import Model
from finesse.components import Cavity, Beamsplitter
from ...util import assert_cavities_equivalent


def mdl_triangular():
    """Triangular cavity model.

    NOTE: this isn't a @pytest.fixture because it's used multiple times in functions below.
    """
    m1 = Beamsplitter("m1", T=500e-6, L=0, Rc=12)
    m2 = Beamsplitter("m2", T=15e-6, L=0)
    m3 = Beamsplitter("m3", T=15e-6, L=0)
    model = Model()
    model.connect(m1.p1, m2.p2, L=6)
    model.connect(m2.p1, m3.p2, L=6)
    model.connect(m3.p1, m1.p2, L=6)
    return model


def test_choice_of_cavity_via_doesnt_matter():
    """Test that changing the via node to another node within the cavity makes no difference to \
    the computed cavity properties."""
    # Cavity without via.
    mdl_ref = mdl_triangular()
    cav_ref = Cavity("cav_ref", mdl_ref.m1.p1.o, mdl_ref.m1.p2.i)
    mdl_ref.add(cav_ref)

    # Cavities with vias.
    mdl_via1 = mdl_triangular()
    cav_via1 = Cavity("cav_via1", mdl_via1.m1.p1.o, via=mdl_via1.m2.p2.i)
    mdl_via1.add(cav_via1)
    mdl_via2 = mdl_triangular()
    cav_via2 = Cavity("cav_via2", mdl_via2.m1.p1.o, via=mdl_via2.m2.p1.o)
    mdl_via2.add(cav_via2)
    mdl_via3 = mdl_triangular()
    cav_via3 = Cavity("cav_via3", mdl_via3.m1.p1.o, via=mdl_via3.m3.p2.i)
    mdl_via3.add(cav_via3)
    mdl_via4 = mdl_triangular()
    cav_via4 = Cavity("cav_via4", mdl_via4.m1.p1.o, via=mdl_via4.m3.p1.o)
    mdl_via4.add(cav_via4)

    assert_cavities_equivalent(cav_ref, cav_via1)
    assert_cavities_equivalent(cav_ref, cav_via2)
    assert_cavities_equivalent(cav_ref, cav_via3)
    assert_cavities_equivalent(cav_ref, cav_via4)


def test_cavity_via_not_in_path_raises_exception():
    """Test that when a via node is specified that's not in the path, an exception is raised.

    See Also
    --------
    :func:`tests.unit.models.tracing.test_paths.test_optical_path_via_node_not_in_path_raises_exception`
    """

    # The used model connects m1.p1.o -> m2.p2.i, m2.p1.o -> m3.p2.i, m3.p1.o -> m1.p2.i.
    with pytest.raises(NetworkXNoPath):
        mdl = mdl_triangular()
        cav = Cavity("cav", mdl.m1.p1.o, via=mdl.m1.p1.i)
        mdl.add(cav)
    with pytest.raises(NetworkXNoPath):
        mdl = mdl_triangular()
        cav = Cavity("cav", mdl.m1.p1.o, via=mdl.m1.p3.i)
        mdl.add(cav)
    with pytest.raises(NetworkXNoPath):
        mdl = mdl_triangular()
        cav = Cavity("cav", mdl.m1.p1.o, via=mdl.m2.p2.o)
        mdl.add(cav)
    with pytest.raises(NetworkXNoPath):
        mdl = mdl_triangular()
        cav = Cavity("cav", mdl.m1.p1.o, via=mdl.m2.p4.o)
        mdl.add(cav)



"""Text utilities."""


def ngettext(n, fsingle, fplural, sub=True):
    """Get the singular or plural form of the specified messages based on n.

    Simplified version of the Python standard library function :func:`gettext.ngettext`.

    Parameters
    ----------
    n : int
        The number to use to decide which form to return.

    fsingle, fplural : str
        Single and plural templates.

    sub : bool, optional
        Substitute `n` into the templates; defaults to True.
    """
    if n == 1:
        return fsingle % n if sub else fsingle
    return fplural % n if sub else fplural


def option_list(sequence, final_sep="or"):
    """Build a list from `sequence` with commas and a final "or"."""
    sequence = list(sequence)

    if len(sequence) <= 1:
        return "".join(sequence)

    if len(sequence) == 2:
        return f"{sequence[0]} {final_sep} {sequence[1]}"

    commasep = ", ".join(sequence[:-1])
    return f"{commasep} {final_sep} {sequence[-1]}"


def format_section(header, body, ruler=True, ruler_char="="):
    """Format text in sections."""
    text = f"{header}\n"

    if ruler:
        text += f"{ruler_char * len(header)}\n"

    if body:
        text += f"\n{body}\n"

    return text


def format_bullet_list(items, indent=4, bullet_char="-"):
    """Format items into a bullet list."""
    pre = " " * indent
    return "\n".join([f"{pre}{bullet_char} {item}" for item in items])


def add_linenos(linenos, lines):
    """Add line numbers to the start of lines.

    Parameters
    ----------
    linenos : sequence of int
        The line numbers, in the same order as `lines`.

    lines : sequence of str
        The lines.

    Returns
    -------
    sequence of str
        The lines with prepended line numbers.
    """
    # Use as many columns as required to fit the largest line number.
    wlinenocol = max([len(str(lineno)) for lineno in linenos])
    return [f"{lineno:>{wlinenocol}}: {line}" for lineno, line in zip(linenos, lines)]


def stringify(item):
    """Recursively stringify `item`.

    This is useful for when it doesn't make sense or isn't possible to override the
    __repr__ method of an object to get a compact string representation.
    """
    if isinstance(item, (list, tuple)):
        return f"[{', '.join(stringify(i) for i in item)}]"
    return str(item)


def stringify_graph_gml(graph):
    """Convert the specified NetworkX graph to string representation using GML
    markup."""
    from io import BytesIO
    import networkx as nx

    graphbytes = BytesIO()
    nx.write_gml(graph, graphbytes, stringify)
    graphbytes.seek(0)

    return graphbytes.read().decode("utf-8")



"""Test utilities.

Put any misc functionality shared by multiple tests here.
"""

import numpy as np
import networkx as nx
from hypothesis import assume
from hypothesis.strategies import (
    composite,
    recursive,
    one_of,
    integers,
    floats,
    complex_numbers,
    characters,
    text,
    lists,
    sampled_from,
    from_regex,
    dictionaries,
)
from finesse import Model
from finesse.symbols import OPERATORS
from finesse.components import Cavity, Gauss
from finesse.detectors.general import Detector
from finesse.script.spec import KatSpec


_SPEC = KatSpec()
_MODEL = Model()

# Names that cannot be used for elements.
RESERVED_NAMES = {*dir(_MODEL), *_SPEC.reserved_names}


# FIXME: don't hard-code these.
BINARY_OPERATIONS = (
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__floordiv__",
    "__pow__",
)


@composite
def numbers(draw, operations=False):
    """Any type of number."""
    cmplx_kwargs = dict(
        min_magnitude=0, max_magnitude=1, allow_nan=False, allow_infinity=False
    )

    if operations:
        # Wrap complex numbers in Operation objects.
        cmplx_strategy = complex_number_operations(**cmplx_kwargs)
    else:
        cmplx_strategy = complex_numbers(**cmplx_kwargs)

    return draw(
        one_of(
            floats(min_value=-1, max_value=1, allow_nan=False, allow_infinity=False),
            integers(min_value=-1, max_value=1),
            cmplx_strategy,
        )
    )


@composite
def complex_number_operations(draw, **kwargs):
    number = draw(complex_numbers(**kwargs))
    operation = OPERATORS["__sub__"] if number.imag < 0 else OPERATORS["__add__"]
    return operation(number.real, abs(number.imag))


@composite
def open_expressions(draw, rhs):
    lval = draw(numbers())
    operation = draw(sampled_from(BINARY_OPERATIONS))
    rval = draw(rhs)

    try:
        assume(operation not in ("/", "//") and rhs != 0)
    except TypeError:
        # Input type cannot be compared to zero.
        pass

    return OPERATORS[operation](lval, rval)


# TODO: make an Operation for parenthesised expressions
# @composite
# def bracketed_expressions(draw, rhs):
#     return f"({draw(open_expressions(rhs))})"


@composite
def expressions(draw, rhs=numbers()):
    # return draw(one_of(open_expressions(rhs), bracketed_expressions(rhs)))
    return draw(open_expressions(rhs))


@composite
def recursive_expressions(draw, max_leaves=25):
    """Generate nested expressions up to max_leaves."""
    return draw(
        recursive(
            expressions(), lambda rhs: expressions(rhs=rhs), max_leaves=max_leaves
        )
    )


@composite
def recursive_arrays(draw, operations=False):
    """Generate nested arrays up to max_leaves."""
    return [draw(recursive(numbers(operations=operations), lists))]


@composite
def line(draw, nonempty=False):
    """A single line of text."""
    # Blacklist control characters.
    return draw(
        text(
            min_size=1 if nonempty else 0,
            alphabet=characters(blacklist_categories=("Cs", "Cc")),
        )
    )


@composite
def kat_name(draw):
    """Valid kat script name."""
    name = draw(from_regex("^[a-zA-Z_][a-zA-Z0-9_]*$", fullmatch=True))
    assume(name not in RESERVED_NAMES)
    return name


@composite
def kat_empty(draw, multiline=False):
    if multiline:
        pattern = "^[ \t]*$"
    else:
        pattern = "^[ \t\n]*$"
    return draw(from_regex(pattern, fullmatch=True))


@composite
def kat_comment(draw):
    # Blacklist control characters.
    comment = draw(line())
    return f"#{comment}"


# Valid normal fraction for reflectivity etc.
KAT_NORMAL_FLOAT = floats(min_value=0, max_value=1, allow_nan=False)


@composite
def kat_scalar(draw, nonnegative=False):
    """Kat script scalar value, including SI suffices."""
    min_value = 0 if nonnegative else None
    value = draw(floats(min_value=min_value, allow_nan=False, allow_infinity=False))
    items = [value]

    # Can't add a suffix for scientific notation.
    if "e" not in str(value):
        suffix = draw(sampled_from("pnumkMGT"))
        items.append(f"{value}{suffix}")

    return draw(sampled_from(items))


@composite
def kat_rtl(draw):
    """R, T, L adding up to 1."""
    R = draw(KAT_NORMAL_FLOAT, label="R")
    T = draw(KAT_NORMAL_FLOAT, label="T")
    L = draw(KAT_NORMAL_FLOAT, label="L")

    # Don't allow R, T and L to be too small.
    if R != 0:
        assume(R > 1e-3)
    if T != 0:
        assume(T > 1e-3)
    if L != 0:
        assume(L > 1e-3)

    total = R + T + L
    assume(total > 0)
    R /= total
    T /= total
    L /= total
    return R, T, L


@composite
def kat_laser(draw):
    directive = draw(sampled_from(_SPEC.elements["laser"].aliases))
    name = draw(kat_name())
    params = draw(
        dictionaries(keys=sampled_from(("P", "f", "phase")), values=kat_scalar())
    )
    paramstr = " ".join([f"{k}={v}" for k, v in params.items()])
    return f"{directive} {name} {paramstr}"


@composite
def kat_mirror(draw):
    directive = draw(sampled_from(_SPEC.elements["mirror"].aliases))
    name = draw(kat_name())
    R, T, L = draw(kat_rtl())
    return f"{directive} {name} R={R} T={T} L={L}"


@composite
def kat_beamsplitter(draw):
    directive = draw(sampled_from(_SPEC.elements["beamsplitter"].aliases))
    name = draw(kat_name())
    R, T, L = draw(kat_rtl())
    return f"{directive} {name} R={R} T={T} L={L}"


@composite
def kat_element(draw):
    return draw(one_of(kat_laser(), kat_mirror(), kat_beamsplitter()))


@composite
def kat_script_line(draw):
    """Generate kat script line."""
    return draw(one_of(kat_empty(), kat_comment(), kat_element()))


def assert_models_equivalent(model_a, model_b):
    """Check :class:`.Model` equality.

    This checks that the two specified models are identical in a shallow way, i.e. they have
    elements with the same properties, equivalent networks, actions, etc.

    This is implemented here and not as `Model.__eq__` because :class:`.Model` must be hashable.
    """
    assert_actions_equivalent(model_a.analysis, model_b.analysis)

    for el_a, el_b in zip(sorted(model_a.elements), sorted(model_b.elements)):
        assert_model_elements_equivalent(model_a.elements[el_a], model_b.elements[el_b])

    assert_networks_equivalent(model_a.network, model_b.network)


def assert_networks_equivalent(network_a, network_b):
    """Check that two :class:`NetworkX graphs <nx.classes.graph.Graph>` are equal."""
    assert nx.is_isomorphic(
        network_a, network_b
    ), f"{network_a!r} is not an isomorph of {network_b!r}"


def assert_model_elements_equivalent(comp_a, comp_b):
    """Check :class:`.ModelElement` equality.

    This is implemented here and not as `ModelElement.__eq__` because `ModelElement`
    must be hashable.
    """

    if not isinstance(comp_a, comp_b.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError(
            f"Cannot compare {comp_a!r} with type {type(comp_a).__name__} to "
            f"{comp_b!r} with type {type(comp_b).__name__}"
        )

    assert comp_a.name == comp_a.name
    for param_a, param_b in zip(comp_a.parameters, comp_b.parameters):
        assert_parameters_equivalent(param_a, param_b)


def assert_parameters_equivalent(param_a, param_b, expected=None):
    """Check model parameter equality.

    This is implemented here and not as `ModelParameter.__eq__` because `ModelParameter`
    must be hashable.
    """
    if not isinstance(param_a, param_b.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError(
            f"Cannot compare {param_a!r} with type {type(param_b).__name__} to "
            f"{param_a!r} with type {type(param_b).__name__}"
        )

    val_a = param_a.eval()
    val_b = param_b.eval()

    assert (
        val_a == val_b
    ), f"{param_a!r} and {param_b!r} do not evaluate to the same value"

    if expected is not None:
        assert val_a == expected, f"{param_a!r} != {expected}"
        assert val_b == expected, f"{param_b!r} != {expected}"


def assert_cavities_equivalent(cav_a, cav_b):
    """Check cavity equality.

    This is implemented here and not as `Cavity.__eq__` because `Cavity` must be
    hashable.
    """
    if not isinstance(cav_b, cav_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(cav_a, cav_b)
    assert isinstance(cav_a, Cavity), f"{cav_a!r} is not a Cavity"
    assert isinstance(cav_b, Cavity), f"{cav_b!r} is not a Cavity"
    assert (
        cav_a.round_trip_length == cav_b.round_trip_length
    ), f"{cav_a!r} round trip length != {cav_b!r} round trip length"
    assert np.all(
        cav_a.gouy == cav_b.gouy
    ), f"{cav_a!r} round trip gouy != {cav_b!r} round trip gouy"
    assert cav_a.FSR == cav_b.FSR, f"{cav_a!r} FSR != {cav_b!r} FSR"
    assert cav_a.loss == cav_b.loss, f"{cav_a!r} loss != {cav_b!r} loss"
    assert cav_a.finesse == cav_b.finesse, f"{cav_a!r} finesse != {cav_b!r} finesse"
    assert cav_a.FWHM == cav_b.FWHM, f"{cav_a!r} FWHM != {cav_b!r} FWHM"
    assert cav_a.pole == cav_b.pole, f"{cav_a!r} pole != {cav_b!r} pole"
    assert np.all(
        cav_a.mode_separation == cav_b.mode_separation
    ), f"{cav_a!r} mode separation != {cav_b!r} mode separation"
    assert np.all(cav_a.S == cav_b.S), f"{cav_a!r} resolution != {cav_b!r} resolution"
    assert np.all(cav_a.ABCDx == cav_b.ABCDx), f"{cav_a!r} ABCDx != {cav_b!r} ABCDx"
    assert np.all(cav_a.ABCDy == cav_b.ABCDy), f"{cav_a!r} ABCDy != {cav_b!r} ABCDy"
    assert np.all(cav_a.q == cav_b.q), f"{cav_a!r} q != {cav_b!r} q"
    assert np.all(cav_a.g == cav_b.g), f"{cav_a!r} g != {cav_b!r} g"


def assert_gauss_commands_equivalent(gauss_a, gauss_b):
    """Check gauss command equality.

    This is implemented here and not as `Gauss.__eq__` because `Gauss` must be hashable.
    """
    if not isinstance(gauss_b, gauss_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(gauss_a, gauss_b)
    assert isinstance(gauss_a, Gauss), f"{gauss_a!r} is not a Gauss"
    assert isinstance(gauss_b, Gauss), f"{gauss_b!r} is not a Gauss"
    # TODO: add more gauss command comparisons.


def assert_detectors_equivalent(det_a, det_b):
    """Check detector equality.

    This is implemented here and not as `Detector.__eq__` because `Detector` must be
    hashable.
    """
    if not isinstance(det_b, det_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(det_a, det_b)
    assert isinstance(det_a, Detector), f"{det_a!r} is not a Detector"
    assert isinstance(det_b, Detector), f"{det_b!r} is not a Detector"
    # TODO: add more detector comparisons.


def assert_actions_equivalent(act_a, act_b):
    """Check action equality.

    This is implemented here and not as `Action.__eq__` because `Action` must be
    hashable.
    """
    # TODO: compare more things.
    # Compare trees.
    # TODO ddb - this doesn't really work for testing equivalence of Actions anymore
    # assert str(act_a) == str(act_b), f"Action {act_a!r} != Action {act_b!r} {str(act_a)} {str(act_b)}"
    return True


def assert_nodes_equivalent(node_a, node_b):
    """Check node equality.

    This is implemented here and not as `Node.__eq__` because `Node` must be hashable.
    """
    assert node_a.name == node_b.name, f"{node_a!r} name != {node_b!r} name"
    assert_model_elements_equivalent(node_a.component, node_b.component)


def assert_generate_reparse_equivalent(parser, generator, model, **kwargs):
    """Helper function for checking that generated kat script for a model reparses into
    an equivalent model."""
    assert (
        model.analysis is not None
    ), "For comparison, model must have an explicit analysis defined."

    script = generator.dump(model, **kwargs)
    parsed_model = parser.parse(script)
    assert_models_equivalent(model, parsed_model)


def assert_trace_trees_equal(t1, t2):
    """Check TraceTree equality."""
    # Both t1 and t2 must be None or not None.
    if t1 is None and t2 is None:
        # Nothing to test.
        return
    elif (t1 is None and t2 is not None) or (t1 is not None and t2 is None):
        assert False

    assert t1.node == t2.node, f"{t1} node {t1.node} != {t2} node {t2.node}"
    assert (
        t1.dependency == t2.dependency
    ), f"{t1} dependency {t1.dependency} != {t2} dependency {t2.dependency}"

    assert_trace_trees_equal(t1.left, t2.left)
    assert_trace_trees_equal(t1.right, t2.right)



import pytest
from finesse.script.containers import KatScript, KatElement
from finesse.script.exceptions import KatScriptError
from ..util import NAME, NUMBER, SPACE
from ..data import SI_PREFICES, INTEGERS, FLOATS, IMAGINARIES


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS + IMAGINARIES)
def test_number(parser, number, _a, _b):
    assert parser.parse(f"fake_element el1 {number}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[NUMBER(1, 18, number)],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


def test_numbers_with_float_exponents_are_invalid(parser):
    """Test that numbers with float exponents are invalid.

    Only integer exponents are allowed by Python's :func:`float`.
    """
    with pytest.raises(KatScriptError):
        parser.parse("fake_element el1 3.141e2.2")


@pytest.mark.parametrize("prefix,exponent", SI_PREFICES.items())
def test_numbers_with_si_prefices(parser, prefix, exponent):
    strval = f"3.141e{exponent}"
    assert parser.parse(f"fake_element el1 {strval}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[NUMBER(1, 18, strval)],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize("prefix", SI_PREFICES)
def test_numbers_with_si_prefices_and_scientific_notation_invalid__first(
    parser, prefix
):
    """Test that numbers with SI prefices and scientific notation together are invalid."""
    with pytest.raises(KatScriptError):
        parser.parse(f"fake_element el1 3.141{prefix}e-6")


@pytest.mark.parametrize("prefix", SI_PREFICES)
def test_numbers_with_scientific_exponents_and_si_prefices_invalid(parser, prefix):
    """Test that numbers with both scientific exponents and SI prefices are invalid."""
    with pytest.raises(KatScriptError):
        parser.parse(f"fake_element el1 3.141e-2{prefix}")


@pytest.mark.parametrize("prefix", SI_PREFICES)
def test_numbers_with_si_prefices_and_scientific_notation_invalid__second(
    parser, prefix
):
    """Test that numbers with SI prefices and scientific notation together are invalid."""
    with pytest.raises(KatScriptError):
        parser.parse(f"fake_element el1 3.141e-6{prefix}")


@pytest.mark.parametrize("fakeprefix", ("mm", "nn", "pp", "z", "o", "l", "r", "i"))
def test_numbers_with_invalid_si_prefices_invalid(parser, fakeprefix):
    """Test that numbers with invalid SI prefices cause an error."""
    # Note: we can't search the error message for exact token because e.g. 'pp' will match 'p'.
    with pytest.raises(KatScriptError):
        parser.parse(f"fake_element el1 3.141{fakeprefix}")



# pylint: disable=no-member

"""Component unit tests."""

import abc
import pytest
import numpy as np
from ... import Faker

FAKE = Faker()

# Invalid component names, i.e. ones containing non-ASCII characters.
NON_ASCII_INVALID = [
    FAKE.lexify("?" * 10, letters=[str(chr(x)) for x in range(1000, 2000)])
    for _ in range(10)
]


class TestComponent(metaclass=abc.ABCMeta):
    """Tests shared by all components.

    Not to be instantiated directly, rather subclassed.
    """

    def test_name__empty_invalid(self, component):
        """Test that empty strings as names are invalid."""
        with pytest.raises(ValueError):
            component(name="")

    @pytest.mark.parametrize("invalid_name", NON_ASCII_INVALID)
    def test_name__non_ascii_invalid(self, component, invalid_name):
        """Test that non-ASCII characters are invalid."""
        with pytest.raises(ValueError):
            component(name=invalid_name)


class TestSurface(TestComponent, metaclass=abc.ABCMeta):
    """Tests shared by all surfaces.

    Not to be instantiated directly, rather subclassed.
    """

    @pytest.mark.parametrize(
        "R,T,L",
        (
            (1, 0, 0),
            (0, 1, 0),
            (0, 0, 1),
            (0.5, 0.5, 0),
            (0.5, 0, 0.5),
            (0, 0.5, 0.5),
            (0.1, 0.1, 0.8),
            (0.9, 0.09, 0.01),
        ),
    )
    def test_rtl(self, component, R, T, L):
        """Test that a surface's R, T and L are correctly set by the constructor."""
        obj = component("cmp1", R=R, T=T, L=L)
        assert float(obj.R) == R
        assert float(obj.T) == T
        assert float(obj.L) == L

    @pytest.mark.skip(
        reason="May need to be moved to integration test on BaseAnaylsis - see #34"
    )
    def test_rtl__two_from_three(self, component):
        """Test that a surface's constructor correctly forces R+T+L = 1 from provided pairs."""

        def _do_two_from_three_test(specified_params, other_param_name):
            obj = component("cmp1", **specified_params)
            assert float(getattr(obj, other_param_name)) == rtl_data[other_param_name]

        rtl_data = dict(R=0.9, T=0.09, L=0.01)
        keys = list(rtl_data)

        data1 = dict(rtl_data)
        del data1[keys[0]]
        _do_two_from_three_test(data1, keys[0])

        data1 = dict(rtl_data)
        del data1[keys[1]]
        _do_two_from_three_test(data1, keys[1])

        data1 = dict(rtl_data)
        del data1[keys[2]]
        _do_two_from_three_test(data1, keys[2])

    @pytest.mark.skip(
        reason="May need to be moved to integration test on BaseAnaylsis - see #34"
    )
    def test_rtl__negative_invalid(self, component):
        """Test that a surface's R, T and L cannot be negative."""
        with pytest.raises(ValueError):
            component("cmp1", R=-1, T=0.5, L=0.5)
        with pytest.raises(ValueError):
            component("cmp1", R=0.5, T=-0.5, L=1)
        with pytest.raises(ValueError):
            component("cmp1", R=0.5, T=0.7, L=-0.2)
        with pytest.raises(ValueError):
            component("cmp1", R=-1, T=-1, L=-1)

    @pytest.mark.skip(
        reason="May need to be moved to integration test on BaseAnaylsis - see #34"
    )
    def test_rc__zero_invalid(self, component):
        """Test that a surface's radius of curvature cannot be 0."""
        with pytest.raises(ValueError):
            component(name="cmp1", R=0.5, T=0.5, Rc=0)

    def test_rc_sets_rcx_and_rcy__single(self, component):
        """Test that setting a surface's Rc to a single value sets Rcx and Rcy to that value."""
        obj = component(name="cmp1")
        obj.Rc = 3.141
        assert np.all(obj.Rc == 3.141)
        assert float(obj.Rcx) == 3.141
        assert float(obj.Rcy) == 3.141

    def test_rc_sets_rcx_and_rcy__separate(self, component):
        """Test that setting a surface's Rc to a two-valued sequence sets Rcx and Rcy to those \
        respective values."""
        obj = component(name="cmp1")
        obj.Rc = (3.141, 6.282)
        part_a, part_b = obj.Rc
        assert (float(part_a), float(part_b)) == (3.141, 6.282)
        assert float(obj.Rcx) == 3.141
        assert float(obj.Rcy) == 6.282



import finesse
import pytest

@pytest.fixture
def siso_laser_amplitude():
    model = finesse.Model()
    model.parse("""
    l l1 P=0.1
    bs bs1 R=0.01 T=0.99
    readout_dc pd1
    amplifier G -10M
    butter A 4 lowpass 70
    link(l1, bs1, pd1, G, A, l1.amp)

    # out of loop witness sensor
    readout_dc witness
    link(bs1, witness)

    noise laser_amp l1.amp.i 1m/&fsig
    noise amplifier G.p2.o &G.gain*(0.2m/&fsig + 0.0001u*&fsig**2)
    noise pd1_dark_noise pd1.DC.o 0.33n

    noise_projection(
        geomspace(1m, 10M, 1000),
        witness.DC
    )
    """)

    return model
    
def test_siso_laser_amplitude(siso_laser_amplitude):
    siso_laser_amplitude.run()
    
def test_siso_laser_amplitude_plot(siso_laser_amplitude):
    siso_laser_amplitude.run().plot()
    siso_laser_amplitude.run().plot("witness.DC")


import click
from .util import (
    parse_path,
    input_file_argument,
    plot_option,
    trace_option,
    legacy_option,
    verbose_option,
    quiet_option,
    fancy_error_option,
    debug_option,
    log_exclude_option,
    network_type_argument,
    graph_layout_argument,
    graphviz_option,
    list_graph_layouts_option,
    plot_graph,
    KatState,
)


@click.command()
@input_file_argument
@plot_option
@trace_option
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def run(
    ctx, input_file, plot, trace, legacy,
):
    """Run a Finesse script."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    try:
        solution = model.run()
    except Exception as error:
        state.print_error(error, title="ANALYSIS ERROR:")

    if trace:
        try:
            trace_solution = model.beam_trace()
        except Exception as error:
            state.print_error(error, title="TRACE ERROR:")

        state.print("Trace:", fg="green", bold=True)
        state.print(trace_solution)

    if solution is not None:
        state.print("Solution:", fg="green", bold=True)
        state.print(solution, indent=1)

        if plot:
            from .. import configure

            configure(plotting=True)
            solution.plot(show=plot)


@click.command()
@input_file_argument
@click.option(
    "--summary/--no-summary",
    is_flag=True,
    default=True,
    show_default=True,
    help="Show summary.",
)
@click.option("--graph/--no-graph", is_flag=True, default=False, help="Show graph.")
@network_type_argument
@graph_layout_argument
@graphviz_option
@list_graph_layouts_option
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def info(ctx, input_file, summary, graph, network_type, layout, graphviz, legacy):
    """Print information about a model parsed from a script."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    if summary:
        state.print("Summary:", fg="green", bold=True)
        state.print(model.info())

    if graph:
        try:
            plot_graph(
                state,
                network=model.get_network(network_type),
                layout=layout,
                graphviz=graphviz,
            )
        except Exception as e:
            state.print_error(e)


@click.command()
@input_file_argument
@click.option(
    "--from", "from_", help="Start node.",
)
@click.option(
    "--to", help="Stop node.",
)
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def path(ctx, input_file, from_, to, legacy):
    """Print information about a path through the parsed model."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    state.print(model.path(from_, to))


@click.command()
@input_file_argument
@click.option(
    "--direction",
    type=click.Choice(("x", "y")),
    help="Direction to compute mismatches in. If not specified, both 'x' and 'y' are computed.",
)
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def mismatches(ctx, input_file, direction, legacy):
    """Mismatch tools."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    try:
        model.print_cavity_mismatches(direction=direction)
    except Exception as error:
        state.print_error(error, title="TRACE ERROR:")


@click.command()
@input_file_argument
@click.option(
    "--from", "from_", help="Start node.",
)
@click.option(
    "--to", help="Stop node.",
)
@click.option(
    "--symbolic",
    is_flag=True,
    default=False,
    show_default=True,
    help="Perform symbolic trace computation.",
)
@click.option(
    "--print-distances",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print component distances matrix.",
)
@click.option(
    "--plot/--no-plot",
    default=True,
    show_default=True,
    help="Display results as figure.",
)
@click.option(
    "--save-figure",
    type=click.File("wb", lazy=False),
    help="Save image of figure to file.",
)
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def trace(
    ctx, input_file, from_, to, symbolic, print_distances, plot, save_figure, legacy
):
    """Tracing tools."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    try:
        trace_solution = model.propagate_beam(from_, to, symbolic=symbolic)
    except Exception as error:
        state.print_error(error, title="TRACE ERROR:")

    state.print("Trace:", fg="green", bold=True)
    state.print(trace_solution)

    if print_distances:
        # FIXME: print using state.
        trace_solution.print_distances_matrix()

    if plot or save_figure:
        from finesse import configure

        configure(plotting=True)
        trace_solution.plot(show=plot, filename=save_figure)


@click.command()
@input_file_argument
@click.option(
    "--from", "from_", help="Start node.",
)
@click.option(
    "--to", help="Stop node.",
)
@click.option(
    "-p",
    "--parameter",
    "params",
    type=(str, click.Choice(("lin", "log")), float, float, int),
    multiple=True,
    help=(
        "Model parameter to animate in the form '<param> lin|log <start> <stop> <step>' (can be "
        "specified multiple times)."
    ),
)
@click.option(
    "--property",
    "props",
    type=click.Choice(("beamsize", "gouy", "curvature", "all")),
    multiple=True,
    help="Property to animate (can be specified multiple times).",
)
@click.option(
    "-i",
    "--interval",
    type=float,
    default=200,
    help="Delay between frames in milliseconds.",
)
@click.option(
    "--save-figure",
    type=click.Path(dir_okay=False, writable=True),
    help="Save image of figure to file.",
)
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def trace_animation(
    ctx, input_file, from_, to, params, props, interval, save_figure, legacy
):
    """Animate a trace."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    model = parse_path(state, input_file, legacy=legacy)

    try:
        trace_solution = model.propagate_beam(from_, to, symbolic=True)
    except Exception as error:
        state.print_error(error, title="TRACE ERROR:")

    def parse_param_sweep(param, scale, start, stop, step):
        import numpy as np

        param = model.reduce_get_attr(param)
        func = np.linspace if scale == "lin" else np.logspace
        return param, func(start, stop, step)

    parameters = dict(parse_param_sweep(*param) for param in params)
    trace_solution.animate(parameters, *props, interval=interval, filename=save_figure)


if __name__ == "__main__":

    @click.group()
    def entry_point():
        pass

    for command in (run, info, path, mismatches, trace, trace_animation):
        entry_point.add_command(command)

    entry_point()



import sys
import io
import networkx
import re

from finesse.components.node import NodeType
from finesse.components import Connector, SignalGenerator, Wire

from functools import reduce


def dfs_edges_to_paths(a, b):
    if b[-1] in ("forward", "nontree"):
        if b[0] == b[1]:
            return a

        if len(a) == 0:
            a.append([b[0]])
        if b[0] != a[-1][-1]:
            a.append([b[0]])

        a[-1].append(b[1])

    return a


def get_loops_network(model):
    """ Reduces a Model network down to an electronic loop network.
    This can be used for drawing diagrams or computing loops if
    needed.
    """
    net = networkx.digraph.DiGraph()

    is_electric = lambda x: all(p.type == NodeType.ELECTRICAL for p in x.ports)
    in_sig = set()
    out_optics = set()
    in_optics = set()

    for el in model.elements.values():
        if isinstance(el, Connector):
            conn_elec_ports = tuple(
                p.type == NodeType.ELECTRICAL for p in el.ports if p.is_connected
            )
            if len(conn_elec_ports) > 1 and all(conn_elec_ports):
                if isinstance(el, SignalGenerator):
                    if is_electric(el.node.component):
                        OUT = el.node.port.full_name
                    else:
                        OUT = "__in_optics_" + el.node.port.full_name
                        in_optics.add(OUT)

                    in_sig.add(el.name)
                    net.add_edge(
                        el.name,
                        OUT,
                        owner=el,
                        port_A=None,
                        port_B=el.node.port.full_name,
                    )
                elif isinstance(el, Wire):
                    A_electric = is_electric(el.portA.component)
                    B_electric = is_electric(el.portB.component)

                    if A_electric:
                        IN = el.portA.component.name
                    else:
                        IN = "__out_optics_" + el.portA.full_name
                        out_optics.add(IN)

                    if B_electric:
                        OUT = el.portB.component.name
                    else:
                        OUT = "__in_optics_" + el.portB.full_name
                        in_optics.add(OUT)

                    net.add_edge(
                        IN,
                        OUT,
                        owner=el,
                        port_A=el.portA.full_name,
                        port_B=el.portB.full_name,
                    )
                else:
                    net.add_node(el.name, owner=el)

    # Change where multiple edges are combined into a summing node
    for n in tuple(net.nodes):
        in_edges = tuple(net.in_edges(n))
        if len(in_edges) > 1:
            nsum = "__sum_" + n
            net.add_node(nsum)
            net.add_edge(nsum, n)
            for edge in in_edges:
                a, b = edge
                net.add_edge(a, nsum)
                net.remove_edge(a, b)

    # optic_couplings = []

    # Could add in optic-optic coupling if needed
    # for o in set(in_optics):
    #     for i in set(out_optics):
    #         optic_couplings.append((o,i))
    #         net.add_edge(o, i)
    return net, in_optics, out_optics, in_sig


def get_loops_blockdiag_code(kat, f=sys.stdout):
    re1 = re.compile(
        "(?P<sum>__sum_)*(?:__(?P<optics_dir>in|out)_optics_)*(?P<name>.*)"
    )

    net, in_optics, out_optics, in_sig = get_loops_network(kat)
    print("{", file=f)
    print("default_shape = roundedbox;", file=f)
    for n in net.nodes:
        res = re1.match(n).groupdict()
        if res is None:
            raise Exception(f"Unexpected {n}")
        if res["sum"] is not None:
            print(f"{n} [ shape = circle , label='+', width=20, height=20];", file=f)
        elif res["optics_dir"] == "out":
            print(f"{n} [ shape = beginpoint, label = '{res['name']}'];", file=f)
        elif res["optics_dir"] == "in":
            print(f"{n} [ shape = endpoint, label = '{res['name']}'];", file=f)
        else:
            print(f"{n} [ width = {40} , label = '{res['name']}'];", file=f)

    print(
        f"""
    # Inputs into the optics model
    group {{
        {" ".join(in_optics)}
    }}
    # Outputs into the optics model
    group {{
        {" ".join(out_optics)}
    }}
    """,
        file=f,
    )
    for o in set(out_optics) | set(in_sig):
        paths = reduce(
            dfs_edges_to_paths, networkx.dfs_labeled_edges(net, source=o), []
        )
        for p in paths:
            print(" -> ".join(p) + ";", file=f)

    print("}", file=f)


def display_blockdiag_output(cell, output_format="svg", return_svg=False):
    import tempfile
    import blockdiag.command
    from IPython.core.displaypub import publish_display_data

    command = blockdiag.command
    mime_type = {
        "png": "image/png",
        "svg": "image/svg+xml",
    }

    with tempfile.NamedTemporaryFile(suffix=".diag", delete=True) as f:
        cell += "\n"
        f.write(cell.encode("utf-8"))
        f.flush()
        with tempfile.NamedTemporaryFile(suffix="." + output_format, delete=True) as p:
            args = ["-T", output_format, "-o", p.name, f.name]
            command.main(args=args)
            p.seek(0)
            data = p.read()

            if output_format in ["svg"]:
                data = data.decode("utf-8")

            publish_display_data({mime_type.get(output_format, "text/plain"): data})

    if return_svg:
        return data


def print_loops_blockdiag(kat):
    get_loops_blockdiag_code(kat)


def display_loops_blockdiag(kat):
    f = io.StringIO()
    get_loops_blockdiag_code(kat, f)
    display_blockdiag_output(f.getvalue())



from hypothesis import given, settings
from finesse.script import parse
from ......util import kat_script_line


@given(line=kat_script_line())
@settings(deadline=1000)  # See #292.
def test_kat_script_line_fuzzing(line):
    parse(line)



"""Beamsplitter unit tests."""

import pytest
import numpy as np
from finesse.components import Beamsplitter
from . import TestSurface


@pytest.fixture
def component():
    """Set the fixture used by TestMirrorSurface to a beamsplitter."""
    return Beamsplitter


class TestBeamsplitterSurface(TestSurface):
    """Test surface properties (R, T, L, Rc, etc.)."""


@pytest.mark.parametrize(
    "Rcx,Rcy",
    (
        (np.inf, np.inf),
        (np.inf, -np.inf),
        (-np.inf, np.inf),
        (-np.inf, -np.inf),
        (1, 1),
        (1, -1),
        (-1, 1),
        (-1, -1),
        (3.141, 6.282),
        (3.141, -6.282),
        (-3.141, 6.282),
        (-3.141, -6.282),
        (1e9, 1e-6),
        (1e9, -1e-6),
        (-1e9, 1e-6),
        (-1e9, -1e-6),
    ),
)
def test_abcd(Rcx, Rcy):
    """Test that the ABCD matrix of particular combinations of ports shows symmetry with other \
    ports.

    The sign of Rc matters when we test against port pairs on the opposite surface. See
    https://finesse.readthedocs.io/en/latest/api/components/mirror/generated/finesse.components.mirror.Mirror.ABCD.html#finesse.components.mirror.Mirror.ABCD.

    For REFLECTION:

            Beamsplitter( Rcx,  Rcy).p1.i -> Beamsplitter( Rcx,  Rcy).p2.o

                                     is equal to:

        1.  Beamsplitter( Rcx,  Rcy).p2.i -> Beamsplitter( Rcx,  Rcy).p1.o
        2.  Beamsplitter(-Rcx, -Rcy).p3.i -> Beamsplitter(-Rcx, -Rcy).p4.o
        3.  Beamsplitter(-Rcx, -Rcy).p4.i -> Beamsplitter(-Rcx, -Rcy).p3.o

    For TRANSMISSION:

            Beamsplitter( Rcx,  Rcy).p1.i -> Beamsplitter( Rcx,  Rcy).p3.o

                                     is equal to:

        1.  Beamsplitter(-Rcx, -Rcy).p3.i -> Beamsplitter(-Rcx, -Rcy).p1.o
        2.  Beamsplitter( Rcx,  Rcy).p2.i -> Beamsplitter( Rcx,  Rcy).p4.o
        3.  Beamsplitter(-Rcx, -Rcy).p4.i -> Beamsplitter(-Rcx, -Rcy).p2.o
    """

    bs_ref = Beamsplitter("bs", Rc=(Rcx, Rcy))
    bs_cmp_p = Beamsplitter("bs", Rc=(Rcx, Rcy))
    bs_cmp_n = Beamsplitter("bs", Rc=(-Rcx, -Rcy))

    ## Each test below checks x- then y-directions.

    # Reflection
    r_lhs_x = bs_ref.ABCD(bs_ref.p1.i, bs_ref.p2.o, direction="x")
    r_lhs_y = bs_ref.ABCD(bs_ref.p1.i, bs_ref.p2.o, direction="y")

    # (1)
    assert np.all(r_lhs_x == bs_cmp_p.ABCD(bs_cmp_p.p2.i, bs_cmp_p.p1.o, direction="x"))
    assert np.all(r_lhs_y == bs_cmp_p.ABCD(bs_cmp_p.p2.i, bs_cmp_p.p1.o, direction="y"))

    # (2)
    assert np.all(r_lhs_x == bs_cmp_n.ABCD(bs_cmp_n.p3.i, bs_cmp_n.p4.o, direction="x"))
    assert np.all(r_lhs_y == bs_cmp_n.ABCD(bs_cmp_n.p3.i, bs_cmp_n.p4.o, direction="y"))

    # (3)
    assert np.all(r_lhs_x == bs_cmp_n.ABCD(bs_cmp_n.p4.i, bs_cmp_n.p3.o, direction="x"))
    assert np.all(r_lhs_y == bs_cmp_n.ABCD(bs_cmp_n.p4.i, bs_cmp_n.p3.o, direction="y"))

    # Transmission
    t_lhs_x = bs_ref.ABCD(bs_ref.p1.i, bs_ref.p3.o, direction="x")
    t_lhs_y = bs_ref.ABCD(bs_ref.p1.i, bs_ref.p3.o, direction="y")

    # (1)
    assert np.all(t_lhs_x == bs_cmp_n.ABCD(bs_cmp_n.p3.i, bs_cmp_n.p1.o, direction="x"))
    assert np.all(t_lhs_y == bs_cmp_n.ABCD(bs_cmp_n.p3.i, bs_cmp_n.p1.o, direction="y"))

    # (2)
    assert np.all(t_lhs_x == bs_cmp_p.ABCD(bs_cmp_p.p2.i, bs_cmp_p.p4.o, direction="x"))
    assert np.all(t_lhs_y == bs_cmp_p.ABCD(bs_cmp_p.p2.i, bs_cmp_p.p4.o, direction="y"))

    # (3)
    assert np.all(t_lhs_x == bs_cmp_n.ABCD(bs_cmp_n.p4.i, bs_cmp_n.p2.o, direction="x"))
    assert np.all(t_lhs_y == bs_cmp_n.ABCD(bs_cmp_n.p4.i, bs_cmp_n.p2.o, direction="y"))



"""Tests for ensuring that ArraySolution.plot produces the expected figures
corresponding to the output data. Note that these do not check that the output
data are correct, only that the produced figures correspond to the expected
figures given these data.
"""

import builtins
import io
import os

import numpy as np
import pytest

from finesse import Model
import finesse.detectors as detectors
from finesse.analysis import xaxis

from matplotlib.testing.decorators import check_figures_equal


### The below is taken from https://stackoverflow.com/a/51742499/
### and is used to remove figure files produced from tests which
### have check_figures_equal applied to them


def patch_open(open_func, files):
    def open_patched(
        path,
        mode="r",
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None,
        closefd=True,
        opener=None,
    ):
        if "w" in mode and not os.path.isfile(path):
            files.append(path)
        return open_func(
            path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
        )

    return open_patched


# All tests in this file do figure creation so set this to autouse
@pytest.fixture(autouse=True)
def cleanup_files(monkeypatch):
    files = []
    monkeypatch.setattr(builtins, "open", patch_open(builtins.open, files))
    monkeypatch.setattr(io, "open", patch_open(io.open, files))
    yield
    for file in files:
        # Modify the above SO answer a bit to ensure that what we're
        # removing is an image file which is produced after
        # check_figures_equal was applied
        if os.path.isfile(file) and str(file).endswith(".png"):
            os.remove(file)


@pytest.fixture
def fp_cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM R=0.9 T=0.1 Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM R=0.9 T=0.1 Rc=2

    cav FP ITM.p2
    """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    return IFO


@check_figures_equal(extensions=("png",))
@pytest.mark.parametrize(
    "change", (("ITM.phi", -90, 90), ("L0.P", 1, 5), ("ETM.xbeta", 1e-9, 1e-6),)
)
def test_scan_single_param__one_subplot(
    fig_test, fig_ref, fp_cavity_model: Model, change
):
    """Test that scanning a single parameter whilst detecting the transmitted power through
    a cavity produces a single figure consisting of one sub-plot with the correct label
    on the x-axis."""
    IFO = fp_cavity_model
    IFO.parse("pd P ETM.p2.o")
    IFO.select_modes(maxtem=2)

    change_name, start, stop = change

    cname, pname = change_name.split(".")
    param = getattr(IFO.elements[cname], pname)

    out = xaxis(param, "lin", start, stop, 100)
    figures = out.plot(
        show=False,
        tight_layout=False,
        _test_fig_handles={detectors.PowerDetector: fig_test},
    )

    fig_ref.subplots().plot(out.x1, out["P"], label="P")
    fig_ref.axes[0].legend()
    fig_ref.axes[0].set_xlabel(f"{cname} {pname} [{param.units}]")
    fig_ref.axes[0].set_ylabel("Power [W]")

    assert figures["P"] is fig_test
    assert figures[detectors.PowerDetector] is fig_test


@check_figures_equal(extensions=("png",))
@pytest.mark.parametrize(
    "change", (("ITM.phi", -90, 90), ("L0.P", 1, 5), ("ETM.xbeta", 1e-9, 1e-6),)
)
def test_scan_single_param__amplitude_phase_subplots(
    fig_test, fig_ref, fp_cavity_model: Model, change
):
    """Test that scanning a single parameter whilst detecting the transmitted field through
    a cavity produces a single figure consisting of two sub-plots, amplitude and phase, with
    the correct label on the x-axis."""
    IFO = fp_cavity_model
    IFO.parse("ad ad00 ETM.p2.o f=0 n=0 m=0")
    IFO.select_modes(maxtem=2)

    change_name, start, stop = change

    cname, pname = change_name.split(".")
    param = getattr(IFO.elements[cname], pname)

    out = xaxis(param, "lin", start, stop, 100)
    figures = out.plot(
        show=False,
        tight_layout=False,
        _test_fig_handles={detectors.AmplitudeDetector: fig_test},
    )

    amplitude_axis = fig_ref.add_subplot(211)
    phase_axis = fig_ref.add_subplot(212, sharex=amplitude_axis)

    amplitude_axis.plot(out.x1, np.abs(out["ad00"]), label="ad00 (abs)")
    amplitude_axis.set_ylabel(r"Amplitude [$\sqrt{W}$]")
    amplitude_axis.legend()
    phase_axis.plot(out.x1, np.angle(out["ad00"], deg=True), label="ad00 (phase)")
    phase_axis.set_ylabel("Phase [deg]")

    phase_axis.set_xlabel(f"{cname} {pname} [{param.units}]")

    assert figures["ad00"] is fig_test
    assert figures[detectors.AmplitudeDetector] is fig_test



"""Fast C functions providing common mathematical routines for the Cython level code in
Finesse.

Note that the functions and classes provided by this library
are generally only for developers or those extending the code
via additional Cython extensions.

Unless otherwise stated, the functions here do **not** work on
:class:`numpy.ndarray` objects. Instead, they operate on scalar
C types as they are primarily intended for use in low-level code.
"""



"""Short file with a couple of tests for checking that model parameters
get reset correctly after axis sweeps are performed."""

import pytest
import numpy as np

from finesse.analysis import xaxis
from finesse import Model


@pytest.fixture
def fp_cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM Rc=2

    cav FP ITM.p2
    """
    )

    return IFO


@pytest.mark.parametrize(
    "change", (("ITM.phi", -90, 90), ("L0.P", 1, 5), ("ETM.xbeta", 1e-9, 1e-6),)
)
def test_non_geometric_parameter_reset__no_symbols(fp_cavity_model: Model, change):
    """Test that non GeometricParameter instances of a model get reset
    correctly after an axis sweep."""
    IFO = fp_cavity_model

    change_name, start, stop = change

    cname, pname = change_name.split(".")
    param = getattr(IFO.elements[cname], pname)

    initial_value = param.value

    xaxis(param, "lin", start, stop, 2)

    assert param.value == initial_value


@pytest.mark.parametrize(
    "change", (("ITM.Rcx", -1.5, -4), ("sc.L", 1, 5), ("ETM.Rcy", 2.4, 8.2),)
)
def test_geometric_parameter_reset__no_symbols(fp_cavity_model: Model, change):
    """Test that GeometricParameter instances of a model get reset
    correctly after an axis sweep, and the associated ABCD matrices
    are re-calculated to the same values too."""
    IFO = fp_cavity_model

    change_name, start, stop = change

    cname, pname = change_name.split(".")
    comp = IFO.elements[cname]
    param = getattr(comp, pname)

    initial_value = param.value
    # Get the direct refs initial numeric ABCD matrices
    initial_numeric_abcds = [M_num for _, M_num in comp._abcd_matrices.values()]

    xaxis(param, "lin", start, stop, 2)

    # Again, get the refs to numeric ABCD matrices
    final_numeric_abcds = [M_num for _, M_num in comp._abcd_matrices.values()]

    assert param.value == initial_value
    for M_num1, M_num2 in zip(initial_numeric_abcds, final_numeric_abcds):
        # First check that the matrix hasn't been re-allocated to a new array
        assert np.shares_memory(M_num1, M_num2)

        # Then check that values are equal
        assert np.all(M_num1 == M_num2)


# TODO (sjr) Add a few tests with symbolics to check that each parameter
#            is reset correctly after an axis sweep



"""Custom exception types raised by different Finesse functions and class methods."""

from .environment import traceback_handler_instance


class FinesseException(Exception):
    """The exception type which gets raised upon a Finesse failure.

    This identifies whether the current session is interactive or not, and consequently
    sets the level of verbosity. This can be overridden by calling
    :func:`~finesse.environment.show_tracebacks` with ``True``.
    """

    def __init__(self, message, **kwargs):
        if not traceback_handler_instance().show_tb:
            head = "\t(use finesse.tb() to see the full traceback)\n"
        else:
            head = "\n"
        message = head + str(message)
        super().__init__(message, **kwargs)

    def _render_traceback_(self):
        """use custom traceback in IPython/Jupyter."""
        tb = traceback_handler_instance()
        tb.store_tb()

        return tb.get_stb()


class ComponentNotConnected(FinesseException):
    pass


class ParameterLocked(FinesseException):
    pass


class NodeException(FinesseException):
    """Exception associated with :class:`.Node` related run-time errors.

    Objects of type `NodeException` store the error message as well as an optional reference to the
    node(s) which caused the exception to be raised.

    Parameters
    ----------
    message : str
        The error message.

    node : :class:`.Node`, optional
        A reference to the offending node(s), defaults to `None`. This can be a single node or a
        sequence of nodes.
    """

    def __init__(self, message, node=None):
        super().__init__(message)
        self.__node = node

    @property
    def node(self):
        """The node(s) responsible for raising this exception instance.

        :getter: Returns the node(s) (either a single :class:`.Node` object
                 or a sequence of these objects) responsible for the exception
                 (read-only).
        """
        return self.__node


class BeamTraceException(FinesseException):
    pass


class TotalReflectionError(FinesseException):
    """Exception indicating total reflection of a beam at a component when performing
    beam tracing.

    Parameters
    ----------
    message : str
        The error message.

    from_node, to_node : :class:`.Node`
        References to the offending source and target nodes, respectively.
    """

    def __init__(self, message, from_node=None, to_node=None):
        super().__init__(message)
        self.__from_node = from_node
        self.__to_node = to_node

    @property
    def coupling(self):
        """The tuple of (from, to) nodes responsible for the total reflection error.

        :getter: Returns the nodes responsible for the exception (read-only).
        """
        return (self.__from_node, self.__to_node)


class ModelAttributeError(FinesseException):
    """Error indicating a model path was not found.

    Model paths can be e.g. `l1.P` or `s1.p1.o`.

    This exists mainly so it can be caught by the parser.
    """

    def __init__(self, pieces):
        self.path = ".".join(pieces)

        if self.path.endswith("."):
            msg = f"'{self.path}' should not end with a '.'"
        else:
            msg = f"model has no attribute '{self.path}'"

        super().__init__(msg)


class ModelParameterDefaultValueError(FinesseException):
    """Error indicating a model element has no default model parameter.

    Some model parameters have defaults, such that they can be referenced in kat script
    using e.g. `&myvar` instead of `&myvar.value`. This error indicates a model element
    without such a default was referenced directly.
    """

    def __init__(self, element):
        super().__init__(
            f"{element.name} cannot be referenced because type '{element.__class__.__name__}' has "
            f"no default model parameter"
        )


class ModelParameterSelfReferenceError(FinesseException):
    """Error indicating a model parameter cannot be set to refer to itself."""

    def __init__(self, value, parameter):
        super().__init__(
            f"cannot set {parameter.full_name} to self-referencing value {value}"
        )
        self.value = value
        self.parameter = parameter



"""
Comparing error signal gradient vs signal TF - should be "identical"
"""
import pytest
import finesse
import numpy as np
from finesse.analysis.actions import FrequencyResponse, Xaxis, Series, Noxaxis


@pytest.fixture
def model_plane_wave():
    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s s1 L0.p1 EOM.p1
        mod EOM f=100M midx=0.1 order=1 mod_type=pm
        s s2 EOM.p2 ITM.p1
        m ITM R=0.9 T=0.1 Rc=-2
        s sCAV ITM.p2 ETM.p1 L=1
        m ETM R=1 T=0 phi=1 Rc=2

        sgen sig ETM.mech.z 1 0

        pd1 REFL_I ITM.p1.o &EOM.f 0
        pd2 REFL_I_fsig ITM.p1.o &EOM.f 0 &fsig.f

        fsig(1)
        xaxis(ETM.phi, lin, -10, 10, 10000)
        """
    )

    return kat


@pytest.fixture
def sol_plane_wave(model_plane_wave):
    return model_plane_wave.run()


@pytest.fixture
def sol_hom(model_plane_wave):
    model_plane_wave.parse(
        """
        cav c1 ITM.p2.o
        modes(maxtem=1)
        """
    )
    model_plane_wave.L0.tem(0, 0, 1, 0)
    model_plane_wave.L0.tem(0, 1, 0.1, 0)

    return model_plane_wave.run()


def test_plane_wave_equal(sol_plane_wave):
    """
    The accuracy of these matching depends on the number of points used.
    """
    sol = sol_plane_wave
    k = 2 * np.pi / 1064e-9
    z = np.deg2rad(sol.x1) / k
    dz = z[1] - z[0]
    grad_res = np.gradient(sol["REFL_I"], dz)
    sig_res = abs(sol["REFL_I_fsig"]) * np.sign(sol["REFL_I_fsig"]).real

    assert all(abs(grad_res - sig_res) / abs(sig_res) < 0.01)


def test_hom_equal(sol_hom):
    """
    The accuracy of these matching depends on the number of points used.
    """
    sol = sol_hom
    k = 2 * np.pi / 1064e-9
    z = np.deg2rad(sol.x1) / k
    dz = z[1] - z[0]
    grad_res = np.gradient(sol["REFL_I"], dz)
    sig_res = abs(sol["REFL_I_fsig"]) * np.sign(sol["REFL_I_fsig"]).real

    assert all(abs(grad_res - sig_res) < 30e3)



############################################################################################
############################################################################################
# Cavity test
############################################################################################
############################################################################################

# reference is the error signal slope, matches with v2
# could probably derive the analytic solution here
reference = -16148063397.725525

@pytest.fixture
def compute_fp_cavity_readout_pd1_slope():
    base = finesse.Model()
    base.parse("""
    variable f1 9099471
    variable nsilica 1.45
    variable Mloss 30u
    variable Larm 3994
    ###########################################################################
    ###   laser
    ###########################################################################
    laser L0 P=125
    mod mod1 f=&f1 midx=0.18 order=1 mod_type=pm
    link(L0, mod1)
    ###########################################################################
    ###   Xarm
    ###########################################################################
    # Distance from beam splitter to X arm input mirror
    s lx1 mod1.p2 ITMXAR.p1
    m ITMXAR R=0 L=20u xbeta=&ITMX.xbeta ybeta=&ITMX.ybeta phi=&ITMX.phi
    s ITMXsub ITMXAR.p2 ITMX.p1 L=0.2 nr=&nsilica
    m ITMX T=0.014 L=&Mloss Rc=-1934
    s LX ITMX.p2 ETMX.p1 L=&Larm
    m ETMX T=5u L=&Mloss Rc=2245 phi=0
    s ETMXsub ETMX.p2 ETMXAR.p1 L=0.2 nr=&nsilica
    m ETMXAR 0 500u xbeta=&ETMX.xbeta ybeta=&ETMX.ybeta phi=&ETMX.phi
    ###########################################################################
    dof DARM ITMX.dofs.z +1 ETMX.dofs.z -1 
    dof HARDP ITMX.dofs.pitch +1 ETMX.dofs.pitch +1 
    dof SOFTP ITMX.dofs.pitch +1 ETMX.dofs.pitch -1
    dof HARDY ITMX.dofs.yaw +1 ETMX.dofs.yaw +1 
    dof SOFTY ITMX.dofs.yaw +1 ETMX.dofs.yaw -1

    readout_rf REFL9 ITMXAR.p1.o f=&f1 phase=90 output_detectors=true

    sgen sig DARM.AC
    pd2 TF REFL9.p1.i f1=&REFL9.f phase1=&REFL9.phase f2=&fsig.f
    """)

    base.fsig.f = 1e-3

    sol = Series(
        FrequencyResponse((base.fsig.f,), "DARM", "REFL9.I", name='tf'),
        Xaxis("DARM.DC", 'lin', -1e-8, 1e-8, 1, relative=True),
        Noxaxis()
    ).run(base)

    k = 2*np.pi/base.lambda0
    TF_gain1 = sol['noxaxis']['TF']
    TF_gain = sol['tf'].out.squeeze()
    error = sol['xaxis']['REFL9_I']
    x = sol['xaxis'].x1
    # convert degree sweep into meters
    slope = np.gradient(error, x[1]-x[0]).mean() *(k*np.rad2deg(1))# W/m

    print("pd1:       ", TF_gain1.real) # units W/m
    print("readout:   ", TF_gain.real) # units W/m
    print("dif errsig:", slope.real)

    return TF_gain1.real, TF_gain.real, slope.real
    
def test_error_signal_slope(compute_fp_cavity_readout_pd1_slope):
    pd1, readout, slope = compute_fp_cavity_readout_pd1_slope
    assert(abs(slope - reference)/abs(reference) < 1e-6)
    
def test_error_pd1(compute_fp_cavity_readout_pd1_slope):
    pd1, readout, slope = compute_fp_cavity_readout_pd1_slope
    assert(abs(pd1 - reference)/abs(reference) < 1e-6)

@pytest.mark.xfail(reason='Weird 0.1\% error in readout method, unsure why')
def test_error_readout(compute_fp_cavity_readout_pd1_slope):
    pd1, readout, slope = compute_fp_cavity_readout_pd1_slope
    assert(abs(readout - reference)/abs(reference) < 1e-6)
    


import pytest
from finesse.script import unparse


@pytest.mark.parametrize(
    "args,kwargs,expected",
    (
        ([], {"maxtem": 0}, "maxtem=0"),
        ([], {"maxtem": 2}, "maxtem=2"),
        (["off"], {}, None),
        (["even"], {"maxtem": 4}, "modes=even, maxtem=4"),
    ),
)
def test_modes(model, args, kwargs, expected):
    model.select_modes(*args, **kwargs)
    script = unparse(model)
    modelines = filter(lambda line: line.startswith("modes("), script.splitlines())

    if expected is not None:
        modeline = next(modelines)
        assert modeline == f"modes({expected})"
    else:
        assert not any(modelines)



"""Test `reraise_in_spec_context` which rewrites parsing errors with more spec-specific
information."""

import re
import pytest
from finesse.script.exceptions import KatScriptError


@pytest.fixture
def compiler(compiler):
    element = lambda instance, a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    cmd = lambda model, a, b: None
    compiler.spec.register_command(
        "fake_command", {"setter": cmd, "getter": cmd, "singular": True},
    )
    analysis = lambda instance, a: None
    compiler.spec.register_analysis(
        "fake_analysis", {"setter": analysis, "getter": analysis}
    )

    return compiler


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element myparam=1",
            "\nline 1: 'fake_element' should be written in the form 'fake_element a b c d'\n"
            "-->1: fake_element myparam=1\n"
            "      ^^^^^^^^^^^^",
            id="element-without-name",
        ),
        pytest.param(
            "fake_command 1",
            "\nline 1: 'fake_command' should be written in the form 'fake_command(a, b)'\n"
            "-->1: fake_command 1\n"
            "      ^^^^^^^^^^^^",
            id="command-without-opening-parenthesis-arg",
        ),
        pytest.param(
            "fake_command even 4",
            "\nline 1: 'fake_command' should be written in the form 'fake_command(a, b)'\n"
            "-->1: fake_command even 4\n"
            "      ^^^^^^^^^^^^",
            id="command-without-opening-parenthesis-args",
        ),
        pytest.param(
            "fake_command myparam=1",
            "\nline 1: 'fake_command' should be written in the form 'fake_command(a, b)'\n"
            "-->1: fake_command myparam=1\n"
            "      ^^^^^^^^^^^^",
            id="command-without-opening-parenthesis-kwarg",
        ),
        pytest.param(
            "fake_analysis myparam=1",
            "\nline 1: 'fake_analysis' should be written in the form 'fake_analysis(a)'\n"
            "-->1: fake_analysis myparam=1\n"
            "      ^^^^^^^^^^^^",
            id="analysis-without-opening-parenthesis",
        ),
        # Even though a name or '(' is missing, the earlier error is thrown.
        pytest.param(
            "__unregistered_directive__ myparam=1",
            "\nline 1: unknown element or function '__unregistered_directive__'\n"
            "-->1: __unregistered_directive__ myparam=1\n"
            "      ^^^^^^^^^^^^^^^^^^^^^^^^^^",
            id="unregistered-directive",
        ),
    ),
)
def test_directive_missing_name_or_parenthesis_syntax_error(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)



import pytest
from ..util import COMMENT, IMPLICITLINEEND, ENDMARKER


@pytest.mark.parametrize(
    "string,expected",
    (
        (
            "# comment",
            [COMMENT(1, 1, "# comment"), IMPLICITLINEEND(1, 10), ENDMARKER(2),],
        ),
        (
            "### comment",
            [COMMENT(1, 1, "### comment"), IMPLICITLINEEND(1, 12), ENDMARKER(2),],
        ),
        (
            "### comment ###",
            [COMMENT(1, 1, "### comment ###"), IMPLICITLINEEND(1, 16), ENDMARKER(2),],
        ),
    ),
)
def test_comment(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected



import pytest
from finesse.script.containers import KatScript, KatElement, KatKwarg
from ..util import SPACE, NAME, NUMBER, EQUALS


@pytest.mark.parametrize(
    "script,element",
    (
        (
            # Element with no arguments.
            "fake_element el1",
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[],
                extra=[SPACE(1, 13)],
            ),
        ),
        (
            # Element with 4 positional arguments.
            "fake_element el1 1 2 3 4",
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    NUMBER(1, 18, "1"),
                    NUMBER(1, 20, "2"),
                    NUMBER(1, 22, "3"),
                    NUMBER(1, 24, "4"),
                ],
                extra=[
                    SPACE(1, 13),
                    SPACE(1, 17),
                    SPACE(1, 19),
                    SPACE(1, 21),
                    SPACE(1, 23),
                ],
            ),
        ),
        (
            # Element with 2 positional arguments and 2 keyword arguments.
            "fake_element el1 1 2 c=3 d=4",
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    NUMBER(1, 18, "1"),
                    NUMBER(1, 20, "2"),
                    KatKwarg(
                        key=NAME(1, 22, "c"),
                        value=NUMBER(1, 24, "3"),
                        extra=[EQUALS(1, 23)],
                    ),
                    KatKwarg(
                        key=NAME(1, 26, "d"),
                        value=NUMBER(1, 28, "4"),
                        extra=[EQUALS(1, 27)],
                    ),
                ],
                extra=[
                    SPACE(1, 13),
                    SPACE(1, 17),
                    SPACE(1, 19),
                    SPACE(1, 21),
                    SPACE(1, 25),
                ],
            ),
        ),
        (
            # Element with 4 keyword arguments.
            "fake_element el1 a=1 b=2 c=3 d=4",
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 18, "a"),
                        value=NUMBER(1, 20, "1"),
                        extra=[EQUALS(1, 19)],
                    ),
                    KatKwarg(
                        key=NAME(1, 22, "b"),
                        value=NUMBER(1, 24, "2"),
                        extra=[EQUALS(1, 23)],
                    ),
                    KatKwarg(
                        key=NAME(1, 26, "c"),
                        value=NUMBER(1, 28, "3"),
                        extra=[EQUALS(1, 27)],
                    ),
                    KatKwarg(
                        key=NAME(1, 30, "d"),
                        value=NUMBER(1, 32, "4"),
                        extra=[EQUALS(1, 31)],
                    ),
                ],
                extra=[
                    SPACE(1, 13),
                    SPACE(1, 17),
                    SPACE(1, 21),
                    SPACE(1, 25),
                    SPACE(1, 29),
                ],
            ),
        ),
    ),
)
def test_element(parser, script, element):
    assert parser.parse(script) == KatScript(arguments=[element], extra=[])



import pytest
from finesse.script import unparse
from finesse.components.laser import Laser


@pytest.fixture
def laser_model(model):
    model.add(Laser("l1"))
    return model


def test_tem_default(laser_model):
    """No tem() command printed when only default is set."""
    script = unparse(laser_model)
    temlines = filter(lambda line: line.startswith("tem("), script.splitlines())
    assert not list(temlines)


@pytest.mark.parametrize(
    "tems",
    (
        # n,m,factor,phase
        ((0, 0, 0, 0),),
        ((0, 0, 0, 0), (1, 0, 1, 0)),
        ((1, 0, 0.5, 45), (0, 1, 0.5, 90)),
    ),
)
def test_tem(laser_model, tems):
    expected_script = []
    for n, m, factor, phase in tems:
        laser_model.l1.tem(n, m, factor, phase)
        expected_script.append(f"tem(l1, n={n}, m={m}, factor={factor}, phase={phase})")

    script = unparse(laser_model)
    temlines = filter(lambda line: line.startswith("tem("), script.splitlines())

    assert set(temlines) == set(expected_script)



"""Integration test cases for detector nodes."""

import pytest
from finesse import Model
from finesse.components import Mirror
from finesse.detectors import PowerDetector


def test_detector_node_cannot_be_added_to_model_when_node_is_not_in_model():
    """Test that a detector cannot have a node not in the same model when added."""
    model_a = Model()
    model_b = Model()

    model_a.add(Mirror("m1"))

    with pytest.raises(Exception):
        model_b.add(PowerDetector("pd1", model_a.m1.p1.o))



"""Optical components performing modulation of beams."""
import logging

import numpy as np

from finesse.components.general import InteractionType, Connector, FrequencyGenerator
from finesse.components.node import NodeType, NodeDirection
from finesse.parameter import (
    float_parameter,
    int_parameter,
    bool_parameter,
    enum_parameter,
    ParameterState,
)

from finesse.enums import ModulatorType

LOGGER = logging.getLogger(__name__)


@float_parameter("f", "Frequency", units="Hz")
@float_parameter("midx", "Modulation index")
@float_parameter("phase", "Phase", units="degrees")
@int_parameter(
    "order",
    "Maximum modulation order",
    changeable_during_simulation=False,
    validate="_check_order",
)
@enum_parameter(
    "mod_type",
    f"Modulation type {tuple(_ for _ in ModulatorType.__members__.keys())}",
    ModulatorType,
    changeable_during_simulation=False,
    validate="_check_mod_type",
)
@bool_parameter("positive_only", "Positive only", changeable_during_simulation=False)
class Modulator(Connector, FrequencyGenerator):
    """Represents a modulator optical component with associated properties such as
    modulation frequency, index and order.

    Parameters
    ----------
    name : str
        Name of newly created modulator.

    f : float or :class:`.Frequency`, optional
        Frequency of the modulation (in Hz) or :class:`.Frequency` object.

    midx : float
        Modulation index, >= 0.

    order : int, optional
        Maximum order of modulations to produce. Must be 1 for
        amplitude modulation. Defaults to 1.

    mod_type : str, optional
        Modulation type, either 'am' (:ref:`amplitude
        modulation<amp_mod>`) or 'pm' (:ref:`phase
        modulation<phase_mod>`). Defaults to 'pm'.

    phase : float, optional
        Relative phase of modulation (in degrees). Defaults to 0.

    positive_only : bool, optional
        If True, only produce positive-frequency sidebands. Defaults to False.
    """

    def __init__(
        self, name, f, midx, order=1, mod_type="pm", phase=0, positive_only=False
    ):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)

        self.f = f
        self.midx = midx
        self.phase = phase
        self.order = order
        self.mod_type = mod_type
        self.positive_only = positive_only

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

        self._register_node_coupling("amp_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("amp_P2o", self.amp.i, self.p2.o)

        self.__freqs_changing = None
        self.__changing_check = set((self.midx, self.phase))

    def _modulation_frequencies(self):
        order = self.order.eval()
        if order - int(order) != 0:
            raise ValueError(
                f"Modulation order for {self.name} must be an integer not {order}"
            )
        order = int(order)
        orders = list(range(-order, 1 + order))
        orders.pop(order)  # remove 0
        fm = np.dot(self.f.ref, orders)
        return fm

    def _couples_frequency(self, ws, connection, f_in, f_out):
        if connection in ("P2i_P1o", "P1i_P2o"):
            if f_in == f_out:
                return True
            elif (f_in, f_out) in ws.carrier_frequency_couplings:
                return True
            elif (f_out, f_in) in ws.carrier_frequency_couplings:
                return True
            elif (f_in, f_out) in ws.signal_frequency_couplings:
                return True
            elif (f_out, f_in) in ws.signal_frequency_couplings:
                return True
            else:
                return False
        else:  # any other signal connection
            return True

    def _check_order(self, value):
        if self.mod_type == ModulatorType.am and int(value) != 1:
            raise ValueError(
                "Modulation order must be 1 when using amplitude modulation"
            )
        return value

    def _check_mod_type(self, value):
        try:
            value = ModulatorType(value)
        except ValueError:
            try:
                value = ModulatorType[value]  # cast input into enum
            except KeyError:
                raise ValueError(
                    f"'{value}' is not a valid Modulator type, options are {tuple(_ for _ in ModulatorType.__members__.keys())}"
                )

        if self.order.state == ParameterState.Numeric:
            if value == ModulatorType.am and self.order > 1:
                raise ValueError("Amplitude modulation can only be used with order 1")
        return value

    @property
    def f(self):
        """Source frequency representing modulation.

        Returns
        -------
        :class:`float` or :class:`.Frequency`
            The modulation frequency.
        """
        return self.__f

    @f.setter
    def f(self, value):
        self.__f = value

    @property
    def order(self):
        """The maximum order of modulations produced by the modulator.

        Returns
        -------
        int
            The maximum modulation order.
        """
        return self.__order

    @order.setter
    def order(self, value):
        self.__order = int(value)

    def _get_workspace(self, sim):
        from finesse.components.modal.modulator import (
            ModulatorWorkspace,
            modulator_carrier_fill,
            modulator_signal_optical_fill,
            modulator_signal_phase_fill,
            modulator_signal_amp_fill,
        )

        ws = ModulatorWorkspace(self, sim, False)

        ws.carrier.add_fill_function(
            modulator_carrier_fill, self.midx.is_changing or self.phase.is_changing
        )

        if sim.signal:
            # modulator for signals doesn't actually need refilling each time if
            # signal frequency is changing, as the coupling depends on the
            ws.signal.add_fill_function(
                modulator_signal_optical_fill,
                self.midx.is_changing or self.phase.is_changing,
            )

            if ws.amp_signal_enabled:
                # Amplitude signal elements do not depend on the signal frequency
                ws.signal.add_fill_function(
                    modulator_signal_amp_fill,
                    self.midx.is_changing or self.phase.is_changing,
                )

            if ws.phs_signal_enabled:
                # Phase signal elements do not depend on the signal frequency
                ws.signal.add_fill_function(
                    modulator_signal_phase_fill,
                    self.midx.is_changing or self.phase.is_changing,
                )
        return ws



from abc import ABC
from math import isclose
import numpy as np

from ..parameter import float_parameter
from ..symbols import Resolving, Symbol
from ..utilities.misc import calltracker
from .general import Connector


@float_parameter("R", "Reflectivity", validate="_check_R", setter="set_RTL")
@float_parameter("T", "Transmission", validate="_check_T", setter="set_RTL")
@float_parameter("L", "Loss", validate="_check_L", setter="set_RTL")
@float_parameter("phi", "Phase", units="degrees")
@float_parameter(
    "Rcx",
    "Radius of curvature (x)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter(
    "Rcy",
    "Radius of curvature (y)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter("xbeta", "Misalignment (x)", units="radians")
@float_parameter("ybeta", "Misalignment (y)", units="radians")
class Surface(ABC, Connector):
    """Abstract optical surface interface providing an object with common properties for
    :class:`.Mirror` and :class:`.Beamsplitter` to inherit from.

    Parameters
    ----------
    name : str
        Name of newly created surface.

    R : float, optional
        Reflectivity of the surface.

    T : float, optional
        Transmissivity of the surface.

    L : float, optional
        Loss of the surface.

    phi : float, optional
        Microscopic tuning of the surface (in degrees).

    Rc : float, optional
        Radius of curvature (in metres); defaults to ``numpy.inf`` to indicate a planar
        surface. An astigmatic surface can be set with `Rc = (Rcx, Rcy)`.
    """

    def __init__(self, name, R, T, L, phi, Rc, xbeta, ybeta):
        Connector.__init__(self, name)

        # only in constructor allow setting of none of R, T, L
        # -> default to equally reflective and tranmissive with no loss
        if R is None and T is None and L is None:
            # Use some default
            self.set_RTL(0.5, T=0.5, L=0)
        else:
            self.set_RTL(R, T, L)

        self.phi = phi
        self.Rc = Rc
        self.xbeta = xbeta
        self.ybeta = ybeta

    def _check_R(self, value):
        if not 0 <= value <= 1:
            raise ValueError("Reflectivity must satisfy 0 <= R <= 1")

        return value

    def _check_T(self, value):
        if not 0 <= value <= 1:
            raise ValueError("Transmissivity must satisfy 0 <= T <= 1")

        return value

    def _check_L(self, value):
        if not 0 <= value <= 1:
            raise ValueError("Loss must satisfy 0 <= L <= 1")

        return value

    def _check_Rc(self, value):
        if value == 0:
            raise ValueError("Radius of curvature must be non-zero.")

        return value

    @calltracker
    def set_RTL(self, R=None, T=None, L=None):
        """Set the values for the R, T and L properties of the surface.

        One of the following combination must be specified:

            - R and T or,
            - R and L or,
            - T and L or,
            - R and T and L

        In the first three cases, the remaining parameter is set via
        the condition,

        .. math::
            R + T + L = 1

        Parameters
        ----------
        R : scalar
            Value of the reflectivity to set.

        T : scalar
            Value of the transmissivity to set.

        L : scalar
            Value of the loss to set.

        Raises
        ------
        ValueError
            If a combination other than one of the above is specified.

            Or if R, T and L are all given but they sum to anything other than one.
        """
        # Try and cast the input into a float, the datatype of the
        # R/T/L parameters. If it can't just ignore it, as it could be
        # None or some callable, or something else. Need this because
        # the usual datatype casting doesn't happen until much later
        # with this setter, and it is used to R/T/L in the contructor
        # not self.R = R, etc.
        if R is not None and not isinstance(R, Symbol):
            R = self.R.datatype_cast(R)
        if T is not None and not isinstance(T, Symbol):
            T = self.T.datatype_cast(T)
        if L is not None and not isinstance(L, Symbol):
            L = self.L.datatype_cast(L)

        # if any of R, T, L set with parameter refs in kat file
        # syntax then just set each attribute directly here and
        # skip completion of third argument / checking of RTL sum
        if any(isinstance(x, Resolving) for x in (R, T, L)):
            if R is not None:
                self.R = R
            if T is not None:
                self.T = T
            if L is not None:
                self.L = L

            return

        N = sum(x is not None for x in (R, T, L))

        if N < 2:
            msg = f"""Invalid combination passed to {self.name}.set_RTL. One of the
following must be specified:

    - R and T or,
    - R and L or,
    - T and L or,
    - R and T and L
            """
            raise ValueError(msg.strip())

        if N == 2:
            old_R = self.R.value
            old_T = self.T.value
            old_L = self.L.value

            try:
                if R is not None:
                    self.R = R
                else:
                    self.R = 1 - (T + L)

                if T is not None:
                    self.T = T
                else:
                    self.T = 1 - (R + L)

                if L is not None:
                    self.L = L
                else:
                    self.L = 1 - (R + T)
            except ValueError:
                self.R = old_R
                self.T = old_T
                self.L = old_L

                raise

        else:
            RTL_sum = R + T + L

            try:
                # Evaluate symbolics, if necessary.
                RTL_sum = RTL_sum.eval()
            except AttributeError:
                pass

            # FIXME: decide what the necessary tolerance is here.
            if not isclose(RTL_sum, 1):
                msg = (
                    f"Expected R + T + L = 1 in {self.name}.set_RTL but "
                    f"got R + T + L = {RTL_sum}"
                )
                raise ValueError(msg)

            self.R = R
            self.T = T
            self.L = L

    @property
    def Rc(self):
        """The radius of curvature of the mirror in metres, for both the tangential and
        sagittal planes.

        :getter: Returns values of both planes' radii of curvature as a
                 :class:`numpy.ndarray` where the first element is the tangential plane
                 RoC and the second element is the sagittal plane RoC.

        :setter: Sets the radius of curvature.

        Examples
        --------
        The following sets the radii of curvature of an object `m`, which
        is a sub-instance of `Surface`, in both directions to 2.5 m:

        >>> m.Rc = 2.5

        Whilst this would set the radius of curvature in the x-direction (tangential
        plane) to 2.5 m and the radius of curvature in the y-direction (sagittal plane)
        to 2.7 m:

        >>> m.Rc = (2.5, 2.7)
        """
        return np.array([self.Rcx.value, self.Rcy.value])

    @Rc.setter
    def Rc(self, value):
        try:
            self.Rcx = value[0]
            self.Rcy = value[1]
        except (IndexError, TypeError):
            self.Rcx = value
            self.Rcy = value

    def actuate_roc(self, dioptres, direction=("x", "y")):
        r"""Actuate on the radius of curvature (RoC) of the surface with a
        specified dioptre shift.

        Sets the RoC to a new value, :math:`R_2`, via,

        .. math::

            R_2 = \frac{2}{d + \frac{2}{R_1}},

        where :math:`R_1` is the current RoC and :math:`d` is the dioptre shift (i.e.
        the `dioptre` argument).

        By default, both planes of curvature are shifted. To shift, e.g., only the
        tangential plane, specify ``direction="x"``.

        Parameters
        ----------
        dioptres : float
            Shift in surface RoC in dioptres.

        direction : tuple or str, optional; default: ("x", "y")
            RoC plane to shift, defaults to both tangential and sagittal.
        """
        rcnew = lambda x: 2 / (dioptres + 2 / x)
        if "x" in direction:
            self.Rcx = rcnew(self.Rcx.value)
        if "y" in direction:
            self.Rcy = rcnew(self.Rcy.value)

    # NOTE this is a bit hacky but gets around using surface.R = value (etc.)
    #      directly in an axis scan without being warned

    def _on_build(self, sim):
        self.set_RTL.__func__.has_been_called = True

    def _on_unbuild(self, sim):
        self.set_RTL.__func__.has_been_called = False



"""Kat compiler expression unit tests."""

import pytest
from finesse.script import parse
from ...util import dedent_multiline
from ..data import (
    EAGER_EXPRESSIONS,
    EAGER_EXPRESSIONS_COMPLEX as _EAGER_EXPRESSIONS_COMPLEX,
    LAZY_EXPRESSIONS,
)


# Only use positive complex numbers for parser tests because the parameter used in the tests (a q
# parameter) doesn't allow negative numbers.
EAGER_EXPRESSIONS_COMPLEX = (
    (script, expected)
    for script, expected in _EAGER_EXPRESSIONS_COMPLEX
    if expected.imag > 0 and expected.imag != float("inf")
)


@pytest.mark.parametrize("expression,expected", EAGER_EXPRESSIONS)
def test_expressions_eager(expression, expected):
    """Expressions that should be eagerly evaluated."""
    model = parse(f"var myvar {expression}")
    assert model.myvar.value.value == expected


@pytest.mark.parametrize("expression,expected", EAGER_EXPRESSIONS_COMPLEX)
def test_expressions_eager_complex(expression, expected):
    """Complex expressions that should be eagerly evaluated."""
    model = parse(
        f"""
        laser l1
        gauss mygauss l1.p1.o q={expression}
        """
    )
    model.mygauss.qx.q == pytest.approx(expected)


@pytest.mark.parametrize("expression,expected", LAZY_EXPRESSIONS)
def test_expressions_lazy(expression, expected):
    """Expressions that should be lazily evaluated."""
    model = parse(f"var myvar {expression}")
    assert model.myvar.value.value == expected


@pytest.mark.parametrize(
    "script,attr,expected",
    (
        (
            dedent_multiline(
                """
                var order 2
                modulator mod1 10M 0.1 1+&order
                """
            ),
            "mod1.order",
            3,
        ),
    ),
)
def test_expressions_with_info_parameter_arguments(script, attr, expected):
    """Test that info parameter arguments can be referenced in expressions.

    See #130 and !57.
    """
    model = parse(script)
    assert model.reduce_get_attr(attr).eval() == expected



from .model import run, info
from .docs import syntax
from .config import config
from .script import convert
from .util import (
    print_banner,
    plot_graph,
    input_file_argument,
    graph_layout_argument,
    graphviz_option,
    verbose_option,
    quiet_option,
    debug_option,
    log_exclude_option,
    KatState,
)

__all__ = (
    "run",
    "info",
    "syntax",
    "config",
    "convert",
    "print_banner",
    "plot_graph",
    "input_file_argument",
    "graph_layout_argument",
    "graphviz_option",
    "verbose_option",
    "quiet_option",
    "debug_option",
    "log_exclude_option",
    "KatState",
)



import re
import pytest
from finesse.script.exceptions import KatSyntaxError
from ...util import dedent_multiline


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "myelement @",
            (
                "\nline 1: illegal character '@'\n"
                "-->1: myelement @\n"
                "                ^"
            ),
            id="illegal-character",
        ),
        pytest.param(
            dedent_multiline(
                """
                laser l1 P=1
                xaxis(l1.f, lin, 10*, 40k, 20)
                """
            ),
            # See #265.
            (
                "\nline 2: syntax error\n"
                "   1: laser l1 P=1\n"
                "-->2: xaxis(l1.f, lin, 10*, 40k, 20)\n"
                "                          ^"
            ),
            id="incomplete-expression",
        ),
        pytest.param(
            "modes ()",
            ("\nline 1: space not allowed here\n" "-->1: modes ()\n" "           ^"),
            id="space-after-function-0-params",
        ),
        pytest.param(
            "modes (maxtem=1)",
            (
                "\nline 1: space not allowed here\n"
                "-->1: modes (maxtem=1)\n"
                "           ^"
            ),
            id="space-after-function-1-param",
        ),
        pytest.param(
            "modes (odd, maxtem=1)",
            (
                "\nline 1: space not allowed here\n"
                "-->1: modes (odd, maxtem=1)\n"
                "           ^"
            ),
            id="space-after-function-2-params",
        ),
        # Too many trailing commas.
        pytest.param(
            "modes(odd, maxtem=1,,)",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1,,)\n"
                "                          ^"
            ),
            id="two-trailing-commas-nospace",
        ),
        pytest.param(
            "modes(odd, maxtem=1 ,,)",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1 ,,)\n"
                "                           ^"
            ),
            id="two-trailing-commas-prespace",
        ),
        pytest.param(
            "modes(odd, maxtem=1, ,)",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1, ,)\n"
                "                           ^"
            ),
            id="two-trailing-commas-innerspace",
        ),
        pytest.param(
            "modes(odd, maxtem=1,, )",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1,, )\n"
                "                          ^"
            ),
            id="two-trailing-commas-postspace",
        ),
        pytest.param(
            "modes(odd, maxtem=1 , , )",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1 , , )\n"
                "                            ^"
            ),
            id="two-trailing-commas-preinnerpostspace",
        ),
        pytest.param(
            "modes(odd, maxtem=1,,,,)",
            (
                "\nline 1: syntax error\n"
                "-->1: modes(odd, maxtem=1,,,,)\n"
                "                          ^"
            ),
            id="four-trailing-commas",
        ),
        pytest.param(
            dedent_multiline(
                """
                modes(
                    odd,
                    maxtem=1,
                    # comment
                    ,
                )
                """
            ),
            (
                "\nline 5: syntax error\n"
                "   4:     # comment\n"
                "-->5:     ,\n"
                "          ^"
            ),
            id="two-trailing-commas-comment-multiline",
        ),
    ),
)
def test_syntax_error(parser, script, error):
    with pytest.raises(KatSyntaxError, match=re.escape(error)):
        parser.parse(script)



"""
Top-level objects which specific optical, and otherwise, components
should inherit from.
"""

from copy import copy
from collections import OrderedDict
import enum
import logging
import numbers

import numpy as np

from finesse.components.node import NodeType, Port
from finesse.components.modal.variable import VariableWorkspace
from finesse.exceptions import ComponentNotConnected
from finesse.parameter import float_parameter
from finesse.utilities import check_name, is_iterable
from finesse.element import ModelElement

from collections import namedtuple

DOFDefinition = namedtuple("DOFDefinition", ("DC", "AC", "DC_2_AC_scaling"))

LOGGER = logging.getLogger(__name__)


def borrows_nodes():
    """Flags that a ModelElement will be making references to nodes
    owner by other elements, or borrows a reference."""

    def func(cls):
        cls._borrows_nodes = True
        return cls

    return func


@enum.unique
class CouplingType(enum.Enum):
    """An enum describing the type of coupling
    between two nodes.
    """

    OPTICAL_TO_OPTICAL = 0
    OPTICAL_TO_ELECTRICAL = 1
    OPTICAL_TO_MECHANICAL = 2
    ELECTRICAL_TO_ELECTRICAL = 3
    ELECTRICAL_TO_OPTICAL = 4
    ELETRICAL_TO_MECHANICAL = 5
    MECHANICAL_TO_MECHANICAL = 6
    MECHANICAL_TO_OPTICAL = 7
    MECHANICAL_TO_ELECTRICAL = 8


@enum.unique
class NoiseType(enum.Enum):
    """An enum describing the type of noise a component generates."""

    QUANTUM = 0


def determine_coupling_type(from_node, to_node):
    """Retrieves the type of coupling (see :class:`.CouplingType`)
    between two nodes.

    Parameters
    ----------
    from_node : :class:`.Node`
        Node which couples into `to_node`.

    to_node : :class:`.Node`
        Node which has a coupling from `from_node`.

    Returns
    -------
    coupling_t : :class:`.CouplingType`
        The type of coupling between the two given nodes.
    """
    convert = {
        (NodeType.OPTICAL, NodeType.OPTICAL): CouplingType.OPTICAL_TO_OPTICAL,
        (NodeType.OPTICAL, NodeType.ELECTRICAL): CouplingType.OPTICAL_TO_ELECTRICAL,
        (NodeType.OPTICAL, NodeType.MECHANICAL): CouplingType.OPTICAL_TO_MECHANICAL,
        (
            NodeType.ELECTRICAL,
            NodeType.ELECTRICAL,
        ): CouplingType.ELECTRICAL_TO_ELECTRICAL,
        (NodeType.ELECTRICAL, NodeType.OPTICAL): CouplingType.ELECTRICAL_TO_OPTICAL,
        (
            NodeType.ELECTRICAL,
            NodeType.MECHANICAL,
        ): CouplingType.ELETRICAL_TO_MECHANICAL,
        (
            NodeType.MECHANICAL,
            NodeType.MECHANICAL,
        ): CouplingType.MECHANICAL_TO_MECHANICAL,
        (NodeType.MECHANICAL, NodeType.OPTICAL): CouplingType.MECHANICAL_TO_OPTICAL,
        (
            NodeType.MECHANICAL,
            NodeType.ELECTRICAL,
        ): CouplingType.MECHANICAL_TO_ELECTRICAL,
    }
    return convert[(from_node.type, to_node.type)]


@enum.unique
class InteractionType(enum.Enum):
    """An enum describing the type of interaction
    between two nodes.
    """

    REFLECTION = 0
    TRANSMISSION = 1


class FrequencyGenerator:
    """
    The base class for components which generate optical frequencies.

    A component inheriting from this class will allow the model to query
    the component to ask what frequencies it wants to use. Frequency generation
    comes in the form of either a laser or something that modulates.
    """

    def _couples_frequency(self, ws, connection, frequency_in, frequency_out):
        """
        This method returns whether this element is coupling frequencies for
        a particular connection.

        For example, a modulator would modulator frequencies when a field
        passes through it. Or a suspended mirror would couple the upper and
        lower sidebands on reflection due to radiation pressure effects.

        Parameters
        ----------
        ws : :class:`.ElementWorkspace`
            Workspace for this particular component
        connection : str
            Name of the connection being queried
        frequency_in : :class:`.Frequency`
            Input frequency
        frequency_out : :class:`.Frequency`
            Output frequency

        Returns
        -------
        bool
            True if frequencies couple at this element
        """
        return False

    def _modulation_frequencies(self):
        return []

    def _source_frequencies(self):
        return []

    def _on_response(self, freqWeakRef):
        """
        Callback function that model calls to say whether requested
        frequency was granted or not. If so a weak reference to the
        frequency object is returned.

        Parameters
        ----------
        freqWeakRef: Weakref.ref
            Weak reference to the assigned Frequency object

        Raises
        ------
        Exception
            If passed a name to a frequency this element did not
            ask for originally.
        """
        # if freqWeakRef() in self.__requested_frequencies:
        self.__frequencies.append(freqWeakRef)
        # else:
        #    raise Exception(f"{self} never requested a frequency {freqWeakRef()}")


class NoiseGenerator:
    """
    The base class for components which generate some kind of noise.

    A component inheriting from this class will allow the model to query
    the component to ask what noise it generates.
    """

    def __init__(self):
        self.__noises = {}

    def _register_noise_output(
        self, name, node, noise_type,
    ):
        if noise_type not in self.__noises.keys():
            self.__noises[noise_type] = []

        self.__noises[noise_type].append((name, node))

    def _couples_noise(self, ws, node, noise_type, frequency_in, frequency_out):
        """
        This method returns whether the noise sidebands this element
        produces are covariant for the specified frequencies at this
        node.

        Parameters
        ----------
        ws : :class:`.ElementWorkspace`
            Workspace for this particular component
        node : :class:`.Node`
            Node to check.
        noise_type : :class:`.NoiseType`
            NoiseType to check.
        frequency_in : :class:`.Frequency`
            Input frequency
        frequency_out : :class:`.Frequency`
            Output frequency

        Returns
        -------
        bool
            True if the specified sidebands are covariant.
        """
        return frequency_in.index == frequency_out.index

    @property
    def noises(self):
        return self.__noises


class Connector(ModelElement):
    """Base class for any component which connects nodes together.

    Internally it stores the nodes and the connections associated with the component. During the
    matrix build this class will then ensure that the matrix elements for each coupling requested
    are allocated and and the required matrix view for editing their values is retrieved.

    The inheriting class should call ``_register_node`` and ``_register_coupling`` to define the
    connections it wants to use.

    Parameters
    ----------
    name : str
        Name of the new `Connector` instance.
    """

    _borrows_nodes = False

    def __init__(self, name):
        super().__init__(name)

        self.__connections = OrderedDict()
        self.__enabled_checks = OrderedDict()
        self.__ports = OrderedDict()
        self.__interaction_types = {}
        self.__nodes = OrderedDict()
        self._abcd_matrices = {}
        self.__abcds_symbolised = False

    def __nodes_of(self, node_type):
        return tuple([node for node in self.nodes.values() if node.type == node_type])

    @property
    def optical_nodes(self):
        """The optical nodes stored by the connector.

        :getter: Returns a list of the stored optical nodes (read-only).
        """
        return self.__nodes_of(NodeType.OPTICAL)

    @property
    def mechanical_nodes(self):
        """The mechanical nodes stored by the connector.

        :getter: Returns a list of the stored mechanical nodes (read-only).
        """
        return self.__nodes_of(NodeType.MECHANICAL)

    @property
    def electrical_nodes(self):
        """The electrical nodes stored by the connector.

        :getter: Returns a list of the stored electric nodes (read-only).
        """
        return self.__nodes_of(NodeType.ELECTRICAL)

    @property
    def ports(self):
        """
        Retrieves the ports available at the object.

        Returns
        -------
        tuple
            Read-only tuple of the ports available at this object.
        """
        return tuple(self.__ports.values())

    def _add_port(self, name, type):
        """
        Creates either an electrical, mechanical, or optical port for this component.
        Each port can then have multiple different nodes associated with - each node
        being an equation to solve for in the linear system.

        For example, the input and output optical field at one part of a component
        would be one port, i.e. reflection from one side of a mirror.

        The port is added to the component object directly to be referenced at a later
        time when the users is connecting components or definining couplings in a component.
        """
        check_name(name)

        if name in self.ports:
            raise Exception("Port %s already exists for this object" % name)

        if hasattr(self, name):
            raise Exception(
                "Port name %s already exists as an attribute in of this object" % name
            )

        p = Port(name, self, type)
        self.__ports[name] = p

        # self._unfreeze()
        assert not hasattr(self, name)
        setattr(self, name, p)
        # self._freeze()

    @property
    def _enabled_checks(self):
        return copy(self.__enabled_checks)

    @property
    def _registered_connections(self):
        return copy(self.__connections)

    @property
    def all_optical_connections(self):
        return {
            k: (self.nodes[v[0]], self.nodes[v[1]])
            for k, v in self.__connections.items()
            if (
                self.coupling_type(self.nodes[v[0]], self.nodes[v[1]])
                == CouplingType.OPTICAL_TO_OPTICAL
            )
        }

    def coupling_type(self, from_node, to_node):
        """Obtains the type of coupling (see :class:`.CouplingType`)
        between the two specified nodes at this component.

        Parameters
        ----------
        from_node : :class:`.Node`
            Node which has a forwards coupling to `to_node`.

        to_node : :class:`.Node`
            Node which has a backwards coupling from `from_node`.

        Returns
        -------
        coupling_t : :class:`.CouplingType`
            The type of coupling between the specified nodes.
        """
        try:
            return determine_coupling_type(from_node, to_node)
        except KeyError:
            return None

    def interaction_type(self, from_node, to_node):
        """Obtains the type of interaction (see :class:`.InteractionType`)
        between the two specified nodes at this component.

        Parameters
        ----------
        from_node : :class:`.Node`
            Node which has a forwards coupling to `to_node`.

        to_node : :class:`.Node`
            Node which has a backwards coupling from `from_node`.

        Returns
        -------
        interaction_t : :class:`.InteractionType`
            The type of interaction between the specified nodes.
        """
        try:
            return self.__interaction_types[(from_node.full_name, to_node.full_name)]
        except KeyError:
            return None

    @property
    def nodes(self):
        """
        All the nodes of all the ports at this component. Order is likely to
        be the order in which the ports and nodes were created, but this is not
        guaranteed.

        Returns
        -------
        nodes : tuple(:class:`.Node`)
        """
        return self.__nodes

    def _register_node_coupling(
        self,
        connection_id: int,
        from_node,
        to_node,
        interaction_type=None,
        forced_name=None,
        enabled_check=None,
    ):
        """
        Registers that this element will connect the output of
        one of its nodes to the input of another.

        Each element is responsible for requesting connections
        to be made within a Model. In practice this means the
        model will allocate the required matrix elements for
        this element to fill in.

        Parameters
        ----------
        connection_id : str
            A unique string ID for this element which is used to identify this connection.

        from_node : :class:`.Node`
            The input node

        to_node : :class:`.Node`
            The output node

        interaction_type : :class:`.InteractionType`, optional
            The type of interaction between the nodes if applicable.

        forced_name : str, optional
            Can force the name of a connection. Used by spaces/wires/joints for
            using the correct name as it doesn't own the node

        enabled_check : function, optional
            A function that returns a True/False if this connection should be enabled
            when a simulation is built from a model. The default value None means True.
        """
        from finesse.components import Space, Wire

        if forced_name is None:
            name = "{}->{}".format(from_node.full_name, to_node.full_name)
        else:
            name = forced_name

        try:
            if self._model is not None:
                raise Exception("Component has already been added to a model")
        except ComponentNotConnected:
            pass

        if name in self.__connections:
            raise Exception(
                "Connection called {} already set at component {}".format(
                    self.name, name
                )
            )

        # TODO: decide whether we actually need this - shouldn't be possible
        #       to connect non-existing ports without python complaining about
        #       an undefined parameter first anyway
        if isinstance(self, Space) or isinstance(self, Wire):
            if from_node.port not in from_node.port.component.ports:
                raise Exception()
            if to_node.port not in to_node.port.component.ports:
                raise Exception()
        else:
            if from_node.full_name not in self.nodes:
                raise Exception(
                    "Node {}.{} is not available at component `{}`".format(
                        from_node.port.name, from_node.name, self.name
                    )
                )
            if to_node.full_name not in self.nodes:
                raise Exception(
                    "Node {}.{} is not available at component `{}`".format(
                        to_node.port.name, to_node.name, self.name
                    )
                )
        if (from_node.full_name, to_node.full_name) in self.__connections.values():
            raise Exception(
                f"Connection between {(from_node.full_name, to_node.full_name)} already exists"
            )

        self.__connections[connection_id] = (from_node.full_name, to_node.full_name)
        if enabled_check:
            self.__enabled_checks[connection_id] = enabled_check

        if interaction_type is not None:
            self.__interaction_types[
                (from_node.full_name, to_node.full_name)
            ] = interaction_type

    def is_valid_coupling(self, from_node, to_node):
        """Flags whether the provided node coupling exists at this connector."""
        return (from_node.full_name, to_node.full_name) in self.__connections.values()

    def check_coupling(self, from_node, to_node):
        """Checks that a coupling exists between `from_node` -> `to_node` and raises
        a ``ValueError`` if not."""
        fname, tname = from_node.full_name, to_node.full_name
        if (fname, tname) not in self.__connections.values():
            raise ValueError(f"No coupling exists between {fname} -> {tname}")

    def _parse_from_to_nodes(self, from_node, to_node):
        if isinstance(from_node, numbers.Integral):
            from_node = getattr(self, f"p{from_node}")
        if isinstance(to_node, numbers.Integral):
            to_node = getattr(self, f"p{to_node}")

        if isinstance(from_node, Port):
            from_node = from_node.i
        if isinstance(to_node, Port):
            to_node = to_node.o

        return from_node, to_node

    def _resymbolise_ABCDs(self):
        # By default components will not have to resymbolise optical ABCD matrices
        pass

    def register_abcd_matrix(self, M_sym, *couplings):
        """Register an ABCD matrix of the given symbolic form for a
        sequence of coupling(s).

        Specifying several couplings for one `M_sym` means that all these
        node couplings will point to the same reference ABCDs --- i.e. the
        matrices kept in the underlying ABCD matrix store will be the same
        blocks of memory.

        .. warning::

            This should only be used in the ``_resymbolise_ABCDs`` method of Connectors,
            when implementing a new component.

        Parameters
        ----------
        M_sym : :class:`numpy.ndarray`
            A 2x2 matrix of symbolic elements describing the analytic form of
            the ABCD matrix for the given coupling(s).

        couplings : sequence of tuples
            Arguments of tuples giving the node couplings which are described by
            the given symbolic ABCD matrix `M_sym`.

            These tuples can be of size two or three, with the first two elements
            always as the `from_node` -> `to_node` instances. The former case implies that
            both the tangential and sagittal plane ABCD matrix couplings are equal
            and so both directions 'x' and 'y' in the underlying matrices store will
            be set to the same values. Whilst the latter case, where the third element is
            either 'x' or 'y', sets just these direction keys to this matrix.
        """
        M_num = np.array(M_sym, dtype=np.float64)

        for coupling in couplings:
            if not is_iterable(coupling) or len(coupling) < 2 or len(coupling) > 3:
                raise ValueError(
                    f"Expected coupling {coupling} passed to _register_abcd_matrix "
                    "to be an iterable of length two (from_node, to_node) or "
                    "length three (from_node, to_node, direction)."
                )

            # No plane given, so same matrix will be used for
            # both planes at this given node coupling
            if len(coupling) == 2:
                from_node, to_node = coupling
                direction = ("x", "y")
            # Otherwise use the specified plane
            else:
                from_node, to_node, direction = coupling
                if direction != "x" and direction != "y":
                    raise ValueError(
                        f"Expected direction argument of coupling {coupling} to "
                        f"be either 'x' or 'y' but got {direction}."
                    )

                direction = (direction,)

            # Check that the node coupling actually exists on this connector
            self.check_coupling(from_node, to_node)

            for d in direction:
                key = (from_node, to_node, d)
                if key in self._abcd_matrices:
                    raise ValueError(
                        f"There is already an ABCD matrix defined for the coupling "
                        f"{from_node.full_name} -> {to_node.full_name} in the "
                        f"plane '{d}'!"
                    )

                LOGGER.info(
                    "For node coupling %s -> %s, in plane %s, registered "
                    "ABCD matrix:\n Symbolic: %s, Current Numeric: %s",
                    from_node.full_name,
                    to_node.full_name,
                    d,
                    M_sym.tolist(),
                    M_num.tolist(),
                )
                self._abcd_matrices[(from_node, to_node, d)] = M_sym, M_num

    def _re_eval_abcds(self):
        for M_sym, M_num in self._abcd_matrices.values():
            M_num[:] = np.array(M_sym, dtype=np.float64)

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
    ):
        """
        Parameters
        ----------
        from_node : :class:`.OpticalNode` or :class:`.Port` or str or int
            Input node. If a port, or string repr of a port, is given then
            the *input* optical node of that port will be used.

        to_node : :class:`.OpticalNode` or :class:`.Port` or str or int
            Output node. If a port, or string repr of a port, is given then
            the *output* optical node of that port will be used.

        direction : str, optional; default: 'x'
            Direction of ABCD matrix computation, default is 'x' for tangential plane.

        symbolic : bool, optional; default: False
            Whether to return the symbolic matrix (as given by equations above). Defaults
            to False such that the numeric matrix is returned.

        copy : bool, optional; default: True
            Whether to return a copy of ABCD matrix (or matrices if `retboth` is true). Defaults
            to True so that the internal matrix cannot be accidentally altered. Use caution
            if switching this flag off.

        retboth : bool, optional; default: False
            Whether to return both the symbolic and numeric matrices as a tuple
            in that order.

        Returns
        -------
        M : :class:`numpy.ndarray`
            The ABCD matrix of the specified coupling for the mirror. This is symbolic
            if either of `symbolic` or `retboth` flags are True.

        M2 : :class:`numpy.ndarray`
            Only returned if `retboth` is True, otherwise just `M` above is returned. This
            will always be the numeric matrix.

        Raises
        ------
        err : :class:`ValueError`
            If no coupling exists between `from_node` and `to_node`.
        """
        from_node, to_node = self._parse_from_to_nodes(from_node, to_node)
        self.check_coupling(from_node, to_node)

        if not self.__abcds_symbolised:
            self._resymbolise_ABCDs()
            self.__abcds_symbolised = True

        key = (from_node, to_node, direction)

        if key not in self._abcd_matrices:
            # If an optical coupling exists and no abcd has been specfied
            # then we can assume it is an identity transformation
            M_sym = np.eye(2, dtype=object)
            M_num = np.eye(2)
        else:
            M_sym, M_num = self._abcd_matrices[key]
            if M_sym is None:
                # Symbolic M is None if an error occurred during the tracing
                # In such a case M_num is a TotalReflectionError or
                # some other exception instance
                raise M_num

            # Evaluate M_sym and assign to memory of M_num so that
            # M_num always corresponds to current parameter state
            M_num[:] = np.array(M_sym, dtype=np.float64)

        return _conn_abcd_return(M_sym, M_num, symbolic, copy, retboth)


def _conn_abcd_return(M_sym, M_num, symbolic: bool, copy: bool, retboth: bool):
    """Helper function for handling symbolic and numeric ABCD matrices in functions."""
    if copy:
        Ms = M_sym.copy()
        Mn = M_num.copy()
    else:
        Ms = M_sym
        Mn = M_num

    if retboth:
        return Ms, Mn

    if symbolic:
        return Ms

    return Mn


@float_parameter("value", "Value", is_default=True)
class Variable(ModelElement):
    def __init__(self, name, value):
        super().__init__(name)
        self.value = value

    def _get_workspace(self, sim):
        return VariableWorkspace(self)
